from __future__ import print_function

import cmd
import datetime
import os
import string
import sys
import textwrap

from cloudmesh_client.common.ConfigDict import ConfigDict
from cloudmesh_client.common.Error import Error
from cloudmesh_client.common.Shell import Shell
from cloudmesh_client.common.util import path_expand
from cloudmesh_client.shell.console import Console
from cloudmesh_client.var import Var

# noinspection PyPep8
import cloudmesh_client
from cloudmesh_client.default import Default
from cloudmesh_client.common.util import get_python
from cloudmesh_client.common.util import check_python
import cloudmesh_client
from cloudmesh_client.common.Printer import Printer
from cloudmesh_client.shell.command import command
from cloudmesh_client.shell.command import PluginCommand
from cloudmesh_client.common.ssh_config import ssh_config
import cloudmesh_client.etc
from cloudmesh_client.cloud.secgroup import SecGroup
from cloudmesh_client.cloud.key import Key

import cloudmesh_client.shell.plugins
from cloudmesh_client.common.StopWatch import StopWatch

from cloudmesh_client.db import CloudmeshDatabase
from cloudmesh_client import setup_yaml

import importlib

cm = CloudmeshDatabase()


class CloudmeshContext(object):
    def __init__(self, **kwargs):
        self.__dict__ = kwargs


PluginCommandClasses = type(
    'CommandProxyClass',
    tuple(PluginCommand.__subclasses__()),
    {})

# print (type(PluginCommand.__subclasses__()))
# print (PluginCommand.__subclasses__())

"""
# not yet implemented
class ConsoleClasses(object):

    def __init__(self, *command_classes):

        classes = []
        for c in command_classes:
            classes.append(PluginCommand.__subclasses__())
        print (classes)

        PluginCommandClasses = type(
            'CommandProxyClass',
            tuple(classes),
            {})

        return PluginCommandClasses
"""


# console = ConsoleFactory(PluginCommand)https://www.rackmountsales.com/kvm-switches/kvm-switch-8-port-combo-usb-ps-2-kvm-switch-sun-imac-compatible-part-kvm-s8.html

# noinspection PyBroadException,PyPep8Naming
class CloudmeshConsole(cmd.Cmd, PluginCommandClasses):
    # class CloudmeshConsole(cmd.Cmd,
    #                       ConsoleClasses(PluginCommand)):

    def precmd(self, line):
        StopWatch.start("command")
        return line

    def postcmd(self, stop, line):
        StopWatch.stop("command")
        if Default.timer:
            print("Timer: {:.4f}s ({})".format(StopWatch.get("command"),
                                               line.strip()))
        return stop

    def onecmd(self, line):
        """Interpret the argument as though it had been typed in response
        to the prompt.

        This may be overridden, but should not normally need to be;
        see the precmd() and postcmd() methods for useful execution hooks.
        The return value is a flag indicating whether interpretation of
        commands by the interpreter should stop.

        """

        line = self.replace_vars(line)

        if line is None:
            return ""
        if line.startswith("!"):
            line.replace("!", "! ")
        line = self.var_replacer(line)
        if line != "hist" and line:
            self._hist += [line.strip()]
        if line.startswith("!") or line.startswith("shell"):
            self.do_shell_exec(line[1:])
            return ""
        cmd, arg, line = self.parseline(line)

        if not line:
            return self.emptyline()
        if os.path.isfile(line):
            self.do_exec(line)
            return ""
        if line.startswith('#') \
                or line.startswith('//') \
                or line.startswith('/*'):
            print(line)
            return self.emptyline()
        if cmd is None:
            return self.default(line)
        self.lastcmd = line
        if line == 'EOF':
            self.lastcmd = ''
        if cmd == '':
            return self.default(line)
        else:
            try:
                func = getattr(self, 'do_' + cmd)
            except AttributeError:
                return self.default(line)
            return func(arg)

    def register_topics(self):
        topics = {}
        for command in PluginCommand.__subclasses__():
            tmp = command.topics.copy()
            topics.update(tmp)
        for name in topics:
            self.register_command_topic(topics[name], name)
        for name in ["q", "EOF", "man", "version", "help", "history",
                     "pause", "quit", "var"]:
            self.register_command_topic("shell", name)

    def __init__(self, context):
        cmd.Cmd.__init__(self)
        self.variables = {}
        self.command_topics = {}
        self.register_topics()
        self.context = context
        # TODO get loglevel from DB or yaml file, if not defined set to ERROR
        self.loglevel = "DEBUG"
        self._hist = []
        if self.context.debug:
            print("init CloudmeshConsole")

        self.prompt = 'cm> '
        self.doc_header = "Documented commands (type help <command>):"
        self.banner = textwrap.dedent("""
            +=======================================================+
            .   ____ _                 _                     _      .
            .  / ___| | ___  _   _  __| |_ __ ___   ___  ___| |__   .
            . | |   | |/ _ \| | | |/ _` | '_ ` _ \ / _ \/ __| '_ \  .
            . | |___| | (_) | |_| | (_| | | | | | |  __/\__ \ | | | .
            .  \____|_|\___/ \__,_|\__,_|_| |_| |_|\___||___/_| |_| .
            +=======================================================+
                                 Cloudmesh Shell
            """)
        # KeyCommands.__init__(self, context)

        #
        # set default cloud and default group if they do not exist
        # use the first cloud in cloudmesh.yaml as default
        #

        Console.set_debug(Default.debug)

        filename = path_expand("~/.cloudmesh/cloudmesh.yaml")
        # moved to import cloudmesh_client

        # create_cloudmesh_yaml(filename)

        setup_yaml()

        # Initialize Logging
        # LogUtil.initialize_logging()

        # sys,exit(1)

        # ##################
        # DEFAULTS
        #

        #
        # SET DEFAULT CLOUD
        #

        value = Default.get(name='cloud', category='general')

        if value is None:

            config = ConfigDict(filename=filename)["cloudmesh"]
            if 'active' in config:
                cloud = config["active"][0]
            else:
                clouds = config["clouds"]
                cloud = list(clouds.keys())[0]
            Default.set('cloud', cloud, category='general')


        #
        # NOT SURE WHAT THIS IS FOR
        #
        value = Default.get(name='default', category='general')
        if value is None:
            Default.set('default', 'default', category='general')

        #
        # SET DEFAULT CLUSTER
        #
        '''
        cluster = ConfigDict(filename="cloudmesh.yaml")["cloudmesh"]["active"][0]

        value = Default.get(name='cluster', category='general')
        if value is None:
            try:
                hosts = ssh_config().names()
                if hosts is not None:
                    cluster = hosts[0]
            except:
                pass  # use the hardcoded cluster

        else:
            cluster = value
        Default.set('cluster', cluster, category='general')
        '''

        #
        # SET DEFAULT GROUP
        #
        group = Default.group
        if group is None:
            Default.set_group("default")

        #
        # LOAD DEFAULTS FROM YAML
        #
        Default.load("cloudmesh.yaml")

        try:
            d = Key.get_from_dir("~/.ssh", store=False)
        except Exception as e:
            Console.error(e.message)

        #
        # SET DEFAULT TIMER
        #
        on = Default.timer


        #
        # SET DEFUALT SECGROUP
        #


        #
        # SET DEFAULT REFRESH
        #
        # r = Default.refresh
        # print ("REFRESH", r)
        # if r is None:
        #     Default.set_refresh("on")

        #
        # SET DEFAULT USER
        #
        user = Default.user
        if user is None:
            user = ConfigDict(filename=filename)["cloudmesh"]["profile"]["user"]
            Default.set_user(user)


        r = Default.secgroup

        if r is None:
            SecGroup.reset_defaults()
            '''
            #secgroup = "{}-default".format(Default.user)
            secgroup = "default"
            Default.set_secgroup(secgroup)
            SecGroup.add_rule_to_db(group=secgroup, name="ssh",from_port="22",to_port="22",protocol="tcp", cidr="0.0.0.0/0")
            SecGroup.add_rule_to_db(group=secgroup, name="http",from_port="80",to_port="80",protocol="tcp", cidr="0.0.0.0/0")
            SecGroup.add_rule_to_db(group=secgroup, name="https", from_port="443", to_port="443", protocol="tcp", cidr="0.0.0.0/0")
            SecGroup.add_rule_to_db(group=secgroup, name="ping", from_port="0", to_port="0", protocol="icmp", cidr="0.0.0.0/0")
            '''

        """
        try:
            sshm = SSHKeyManager()
            m = sshm.get_from_yaml(
                load_order="~/.cloudmesh/cloudmesh.yaml")
            d = dict(m.__keys__)


            sshdb = SSHKeyDBManager()

            for keyname in m.__keys__:
                filename = m[keyname]["path"]
                try:
                    sshdb.add(filename,
                              keyname,
                              source="yaml",
                              uri="file://" + filename)
                except Exception as e:
                    pass
        except Exception as e:
            Console.error("Problem adding keys from yaml file")
        """

        for c in CloudmeshConsole.__bases__[1:]:
            # noinspection PyArgumentList
            c.__init__(self, context)

    def preloop(self):
        """adds the banner to the preloop"""

        if self.context.splash:
            lines = textwrap.dedent(self.banner).split("\n")
            for line in lines:
                # Console.cprint("BLUE", "", line)
                print(line)

    # noinspection PyUnusedLocal
    def do_EOF(self, args):
        """
        ::

            Usage:
                EOF

            Description:
                Command to the shell to terminate reading a script.
        """
        return True

    # noinspection PyUnusedLocal
    def do_quit(self, args):
        """
        ::

            Usage:
                quit

            Description:
                Action to be performed whne quit is typed
        """
        return True

    do_q = do_quit

    def emptyline(self):
        return

    def load_instancemethod(self, location):
        module_name, class_name = location.rsplit(".", 1)
        f = getattr(importlib.import_module(module_name), class_name)
        setattr(self, f.__name__, f)

    # noinspection PyUnusedLocal
    def do_context(self, args):
        """
        ::

            Usage:
                context

            Description:
                Lists the context variables and their values
        """
        """
        :param args:
        :return:
        """
        print(self.context.__dict__)

    # noinspection PyUnusedLocal
    @command
    def do_version(self, args, arguments):
        """
        Usage:
           version [--format=FORMAT] [--check=CHECK]

        Options:
            --format=FORMAT  the format to print the versions in [default: table]
            --check=CHECK    boolean tp conduct an additional check [default: True]

        Description:
            Prints out the version number
        """

        python_version, pip_version = get_python()

        try:
            git_hash_version = Shell.execute('git', 'log -1 --format=%h', traceflag=False, witherror=False)
        except:
            git_hash_version = 'N/A'

        versions = {
            "cloudmesh_client": {
                "name": "cloudmesh_client",
                "version": str(cloudmesh_client.__version__)
            },
            # "cloudmesh_base": {
            #     "name": "cloudmesh_base",
            #     "version": str(cloudmesh_base.__version__)
            # },
            "python": {
                "name": "python",
                "version": str(python_version)
            },
            "pip": {
                "name": "pip",
                "version": str(pip_version)
            },
            "git": {
                "name": "git hash",
                "version": str(git_hash_version)
            }

        }

        print(Printer.write(versions, output=arguments["--format"],
                            order=["name", "version"]))
        if arguments["--check"] in ["True"]:
            check_python()

    def register_command_topic(self, topic, command_name):
        try:
            tmp = self.command_topics[topic]
        except:
            self.command_topics[topic] = []
        self.command_topics[topic].append(command_name)

    def do_help(self, arg):
        """
        ::

            Usage:
                help
                help COMMAND

            Description:
                List available commands with "help" or detailed help with
                "help COMMAND"."""

        if arg:
            try:
                func = getattr(self, 'help_' + arg)
            except AttributeError:
                try:
                    doc = getattr(self, 'do_' + arg).__doc__
                    if doc:
                        self.stdout.write("%s\n" % str(doc))
                        return
                except AttributeError:
                    pass
                self.stdout.write("%s\n" % str(self.nohelp % (arg,)))
                return
            func()
        else:
            names = self.get_names()
            cmds_doc = []
            cmds_undoc = []
            help_page = {}
            for name in names:
                if name[:5] == 'help_':
                    help_page[name[5:]] = 1
            names.sort()
            # There can be duplicates if routines overridden
            prevname = ''
            for name in names:
                if name[:3] == 'do_':
                    if name == prevname:
                        continue
                    prevname = name
                    cmd = name[3:]
                    if cmd in help_page:
                        cmds_doc.append(cmd)
                        del help_page[cmd]
                    elif getattr(self, name).__doc__:
                        cmds_doc.append(cmd)
                    else:
                        cmds_undoc.append(cmd)

            self.stdout.write("%s\n" % str(self.doc_leader))
            self.print_topics(self.doc_header, cmds_doc, 15, 80)
            self.print_topics(self.misc_header, list(help_page.keys()), 15, 80)
            self.print_topics(self.undoc_header, cmds_undoc, 15, 80)

            for topic in self.command_topics:
                topic_cmds = sorted(self.command_topics[topic], key=str.lower)
                self.print_topics(string.capwords(topic + " commands"),
                                  topic_cmds, 15, 80)

    def help_help(self):
        """
        ::

            Usage:
               help NAME

            Prints out the help message for a given function
        """
        print(textwrap.dedent(self.help_help.__doc__))

    def do_exec(self, filename):
        """
        ::

            Usage:
               exec FILENAME

            executes the commands in the file. See also the script command.

            Arguments:
              FILENAME   The name of the file
        """
        if not filename:
            Console.error("the command requires a filename as parameter")
            return

        if os.path.exists(filename):
            with open(filename, "r") as f:
                for line in f:
                    if self.context.echo:
                        Console.ok("cm> {:}".format(str(line)))
                    self.precmd(line)
                    stop = self.onecmd(line)
                    self.postcmd(stop, line)
        else:
            Console.error('file "{:}" does not exist.'.format(filename))
            sys.exit()

    # noinspection PyUnusedLocal
    def do_shell_exec(self, args):
        # just ignore arguments and pass on args
        command = path_expand(args)
        try:
            os.system(command)
        except Exception as e:
            Console.error(e.msg, traceflag=False)

    # noinspection PyUnusedLocal
    # BUG: this does not for some reason execute the arguments
    @command
    def do_shell(self, args, arguments):
        """
        Usage:
           shell ARGUMENTS...

        Description:
            Executes a shell command
        """
        # just ignore arguments and pass on args
        command = path_expand(args)
        try:
            os.system(command)
        except Exception as e:
            Console.error(e, traceflag=False)

    #
    # VAR
    #

    def update_time(self):
        time = datetime.datetime.now().strftime("%H:%M:%S")
        date = datetime.datetime.now().strftime("%Y-%m-%d")

        self.variables['time'] = time
        self.variables['date'] = date

    def var_finder(self, line, c='$'):

        line = line.replace('$', ' $').strip()
        words = line.replace('-', ' ').replace('_', ' ').split(" ")

        variables = []
        for word in words:
            if word.startswith('$'):
                variables.append(word)

        vars = {
            "normal": [],
            "os": [],
            "dot": []
        }
        for word in variables:
            word = word.replace('$', "")
            if word.startswith('os.'):
                vars["os"].append(word)
            elif '.' in word:
                vars["dot"].append(word)
            else:
                vars["normal"].append(word)

        return vars

    def var_replacer(self, line, c='$'):

        vars = self.var_finder(line, c=c)

        for v in vars["normal"]:
            value = str(Var.get(v))
            line = line.replace(c + v, value)
            # replace in line the variable $v with value
        for v in vars["os"]:
            name = v.replace('os.', '')
            if name in os.environ:
                value = os.environ[name]
                line = line.replace(c + v, value)
            else:
                Console.error("can not find environment variable {}".format(
                    v))
                if c + v in line:
                    value = os.environ(v)
                    # replace in line the variable $v with value

        for v in vars["dot"]:
            try:
                config = ConfigDict("cloudmesh.yaml")
                print(config["cloudmesh.profile"])
                value = config[v]
                line = line.replace(c + v, value)
            except Exception as e:
                Console.error("can not find variable {} in cloudmesh.yaml".format(value))
        return line

    def replace_vars(self, line):

        self.update_time()

        newline = line

        variables = Var.list(output='dict')

        if len(variables) is not None:
            for v in variables:
                v = str(v)
                name = str(variables[v]["name"])
                value = str(variables[v]["value"])
                newline = newline.replace("$" + name, value)

        # for v in os.environ:
        #    newline = newline.replace("$" + v.name, os.environ[v])
        newline = path_expand(newline)
        return newline

    '''
    def _add_variable(self, name, value):
        self.variables[name] = value
        # self._list_variables()

    def _delete_variable(self, name):
        self._list_variables()
        del self.variables[name]
        # self._list_variables()

    def _list_variables(self):
        Console.ok(10 * "-")
        Console.ok('Variables')
        Console.ok(10 * "-")
        for v in self.variables:
            Console.ok("{:} = {:}".format(v, self.variables[v]))
    '''

    @command
    def do_var(self, arg, arguments):
        """
        Usage:
            var list
            var delete NAMES
            var NAME=VALUE
            var NAME

        Arguments:
            NAME    Name of the variable
            NAMES   Names of the variable separated by spaces
            VALUE   VALUE to be assigned

        special vars date and time are defined
        """

        if arguments['list'] or arg == '' or arg is None:
            # self._list_variables()
            print(Var.list())
            return ""
        elif arguments['NAME=VALUE'] and "=" in arguments["NAME=VALUE"]:
            (variable, value) = arg.split('=', 1)
            if value == "time" or value == "now":
                value = datetime.datetime.now().strftime("%H:%M:%S")
            elif value == "date":
                value = datetime.datetime.now().strftime("%Y-%m-%d")
            elif value.startswith("default."):
                name = value.replace("default.", "")
                value = Default.get(name=name, category="general")
            elif "." in value:
                try:
                    config = ConfigDict("cloudmesh.yaml")
                    value = config[value]
                except Exception as e:
                    Console.error("can not find variable {} in cloudmesh.yaml".format(value))
                    value = None
            # self._add_variable(variable, value)
            Var.set(variable, value)
            return ""
        elif arguments['NAME=VALUE'] and "=" not in arguments["NAME=VALUE"]:
            try:
                v = arguments['NAME=VALUE']
                # Console.ok(str(self.variables[v]))
                Console.ok(str(Var.get(v)))
            except:
                Console.error('variable {:} not defined'.format(arguments['NAME=VALUE']))

        elif arg.startswith('delete'):
            variable = arg.split(' ')[1]
            Var.delete(variable)
            # self._delete_variable(variable)
            return ""

    @command
    def do_history(self, args, arguments):
        """
        Usage:
            history
            history list
            history last
            history ID
        """
        try:
            if arguments["list"] or args == "":
                print("LIST")
                h = 0
                for line in self._hist:
                    print("{}: {}".format(h, self._hist[h]))
                    h += 1
                return ""

            elif arguments["last"]:
                h = len(self._hist)
                if h > 1:
                    command = self._hist[h - 2]
                    self.precmd(command)
                    stop = self.onecmd(command)
                    self.postcmd(stop, command)
                return ""

            elif arguments["ID"]:
                h = int(arguments["ID"])
                if h in range(0, len(self._hist)):
                    print("{}".format(self._hist[h]))
                    if not args.startswith("history"):
                        command = self._hist[h]
                        self.precmd(command)
                        stop = self.onecmd(command)
                        self.postcmd(stop, command)

                return ""
        except:
            Console.error("could not execute the last command")

    do_h = do_history




def simple():
    context = CloudmeshContext(debug=False,
                               splash=True)
    con = CloudmeshConsole(context)
    con.cmdloop()


# noinspection PyBroadException
def main():
    """cm.

    Usage:
      cm --help
      cm [--echo] [--debug] [--nosplash] [-i] [COMMAND ...]

    Arguments:
      COMMAND                  A command to be executed

    Options:
      --file=SCRIPT  -f  SCRIPT  Executes the script
      -i                 After start keep the shell interactive,
                         otherwise quit [default: False]
      --nosplash    do not show the banner [default: False]
    """

    def manual():
        print(main.__doc__)

    args = sys.argv[1:]

    arguments = {
        '--echo': '--echo' in args,
        '--help': '--help' in args,
        '--debug': '--debug' in args,
        '--nosplash': '--nosplash' in args,
        '-i': '-i' in args}

    echo = arguments["--echo"]
    if arguments['--help']:
        manual()
        sys.exit()

    for a in args:
        if a in arguments:
            args.remove(a)

    arguments['COMMAND'] = [' '.join(args)]

    commands = arguments["COMMAND"]
    if len(commands) > 0:
        if ".cm" in commands[0]:
            arguments["SCRIPT"] = commands[0]
            commands = commands[1:]
        else:
            arguments["SCRIPT"] = None

        arguments["COMMAND"] = ' '.join(commands)
        if arguments["COMMAND"] == '':
            arguments["COMMAND"] = None

    # noinspection PySimplifyBooleanCheck
    if arguments['COMMAND'] == []:
        arguments['COMMAND'] = None

    splash = not arguments['--nosplash']
    debug = arguments['--debug']
    interactive = arguments['-i']
    script = arguments["SCRIPT"]
    command = arguments["COMMAND"]

    context = CloudmeshContext(
        interactive=interactive,
        debug=debug,
        echo=echo,
        splash=splash)
    cmd = CloudmeshConsole(context)

    if script is not None:
        cmd.do_exec(script)

    try:
        if echo:
            print("cm>", command)
        if command is not None:
            cmd.precmd(command)
            stop = cmd.onecmd(command)
            cmd.postcmd(stop, command)
    except Exception as e:
        print("ERROR: executing command '{0}'".format(command))
        print(70 * "=")
        print(e)
        print(70 * "=")
        Error.traceback()

    if interactive or (command is None and script is None):
        cmd.cmdloop()


if __name__ == "__main__":
    main()
    # simple()



from __future__ import print_function
from cloudmesh_client.shell.command import command, PluginCommand, \
    ShellPluginCommand


# noinspection PyPep8Naming
class cm_shell_status(PluginCommand, ShellPluginCommand):
    topics = {"status": "cloud"}

    def __init__(self, context):
        self.context = context
        if self.context.debug:
            print("init status default")

    # noinspection PyUnusedLocal
    @command
    def do_status(self, args, arguments):
        """
        ::

          Usage:
              status
              status db
              status CLOUDS...




          Arguments:



          Options:



        """
        # pprint(arguments)

        if arguments['db']:
            print('status db')
        elif arguments['CLOUDS']:
            print('status CLOUDS...')
        else:
            print('status')



from __future__ import print_function
import textwrap

from cloudmesh_client.shell.command import command
from cloudmesh_client.shell.command import PluginCommand, ShellPluginCommand, \
    CometPluginCommand


# noinspection PyBroadException
class ManCommand(PluginCommand, ShellPluginCommand, CometPluginCommand):
    topics = {"man": "shell"}

    def __init__(self, context):
        self.context = context
        if self.context.debug:
            print("init command man")

    def _print_rst(self, what):
        """
        prints the rst page of the command what

        :param what: the command
        :type what: string

        """

        print
        print("Command - %s::" % what)
        h = None
        exec("h = self.do_%s.__doc__" % what)
        # noinspection PyUnboundLocalVariable
        h = textwrap.dedent(h).replace("::\n\n", "")
        h = textwrap.dedent(h).replace("\n", "\n    ")
        print(h)

    # noinspection PyUnusedLocal
    @command
    def do_man(self, args, arguments):
        """
        ::
        
            Usage:
                   man COMMAND
                   man [--noheader]

            Options:
                   --norule   no rst header

            Arguments:
                   COMMAND   the command to be printed 

            Description:
                man 
                    Prints out the help pages

                man COMMAND
                    Prints out the help page for a specific command
        """
        if arguments['COMMAND'] is None:

            print
            print("Commands")
            print(70 * "=")

            commands = [k for k in dir(self) if k.startswith("do_")]
            commands = sorted(commands, key=str.lower)

        else:
            print(arguments)
            commands = [arguments['COMMAND']]

        for command in commands:
            what = command.replace("do_", "")
            try:
                if not arguments["--noheader"]:
                    print(what)
                    print(70 * "-")
                self._print_rst(what)
            except:
                print("\n    Command documentation %s missing, help_%s" % (what, what))
            print
        return ""



#! /usr/bin/env python

import os

from cloudmesh_client.shell.command import command
from cloudmesh_client.common.ConfigDict import Config
from cloudmesh_client.shell.command import PluginCommand, CloudPluginCommand


class ServerCommand(PluginCommand, CloudPluginCommand):
    topics = {"server": "cloud"}

    def __init__(self, context):
        self.context = context
        if self.context.debug:
            print("init command server")

    # noinspection PyUnusedLocal
    @command
    def do_server(self, args, arguments):
        """
        Usage:
            server

        Options:
          -h --help
          -v       verbose mode

        Description:
          Starts up a REST service and a WEB GUI so one can browse the data in an
          existing cloudmesh database.

          The location of the database is supposed to be in

            ~/.cloud,esh/cloudmesh.db

        """

        # import warnings
        # with warnings.catch_warnings():
        #    warnings.filter("ignore")
        # ignore "SQLALCHEMY_TRACK_MODIFICATIONS")

        from sandman import app
        from sandman.model import activate

        filename = "sqlite:///{}".format(Config.path_expand(
            os.path.join("~", ".cloudmesh", "cloudmesh.db")))

        print("database: {}".format(filename))

        app.config['SQLALCHEMY_DATABASE_URI'] = filename
        app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = False

        activate()

        app.run()




from code import InteractiveConsole
import sys

from cloudmesh_client.shell.command import PluginCommand, ShellPluginCommand, \
    CometPluginCommand
""" This code has been copied and modified from cmd2 to work with cmd3"""


class EmbeddedConsoleExit(SystemExit):
    pass


class Statekeeper(object):

    def __init__(self, obj, attribs):
        self.obj = obj
        self.attribs = attribs
        if self.obj:
            self.save()

    def save(self):
        for attrib in self.attribs:
            setattr(self, attrib, getattr(self.obj, attrib))

    def restore(self):
        if self.obj:
            for attrib in self.attribs:
                setattr(self.obj, attrib, getattr(self, attrib))


# noinspection PyUnresolvedReferences
class PyCommand(PluginCommand, ShellPluginCommand, CometPluginCommand):

    pystate = {}
    locals_in_py = True

    topics = {"open": "shell"}

    def __init__(self, context):
        self.context = context
        if self.context.debug:
            print("init command browser")
        self.locals_in_py = True
        self.pystate['self'] = self

    def do_py(self, arg):
        """
        ::

            Usage:
                py
                py COMMAND

            Arguments:
                COMMAND   the command to be executed

            Description:

                The command without a parameter will be executed and the
                interactive python mode is entered. The python mode can be
                ended with ``Ctrl-D`` (Unix) / ``Ctrl-Z`` (Windows),
                ``quit()``,'`exit()``. Non-python commands can be issued with
                ``cmd("your command")``.  If the python code is located in an
                external file it can be run with ``run("filename.py")``.

                In case a COMMAND is provided it will be executed and the
                python interpreter will return to the command shell.

                This code is copied from Cmd2.
        """
        self.pystate['self'] = self
        arg = arg.strip()
        localvars = (self.locals_in_py and self.pystate) or {}
        interp = InteractiveConsole(locals=localvars)
        interp.runcode('import sys, os;sys.path.insert(0, os.getcwd())')
        if arg:
            interp.runcode(arg)
        else:
            def quit():
                raise EmbeddedConsoleExit

            def onecmd(arg):
                return self.onecmd(arg + '\n')

            def run(arg):
                try:
                    f = open(arg)
                    interp.runcode(f.read())
                    f.close()
                except IOError as e:
                    self.perror(e)
            self.pystate['quit'] = quit
            self.pystate['exit'] = quit
            self.pystate['cmd'] = onecmd
            self.pystate['run'] = run
            try:
                cprt = 'Type "help", "copyright", "credits" or "license" for more information.'
                keepstate = Statekeeper(sys, ('stdin', 'stdout'))
                sys.stdout = self.stdout
                sys.stdin = self.stdin
                interp.interact(banner="Python %s on %s\n%s\n(%s)\n%s" %
                                (sys.version, sys.platform, cprt, self.__class__.__name__, self.do_py.__doc__))
            except EmbeddedConsoleExit:
                pass
            keepstate.restore()



from __future__ import print_function
from pprint import pprint
import collections


def key_prefix_replace(d, prefix, new_prefix=""):
    """
    replaces the list of prefix in keys of a flattened dict

    :param d: the flattened dict
    :param prefix: a list of prefixes that are replaced with a new prefix. Typically this will be ""
    :type prefix: list of str
    :param new_prefix: The new prefix. By default it is set to ""
    :return: the dict with the keys replaced as specified
    """
    items = []
    for k, v in d.items():
        new_key = k
        for p in prefix:
            new_key = new_key.replace(p, new_prefix, 1)
        items.append((new_key, v))
    return dict(items)


def flatten(d, parent_key='', sep='__'):
    """
    flattens the dict into a one dimensional dictionary

    :param d: multidimensional dict
    :param parent_key: replaces from the parent key
    :param sep: the separation character used when fattening. the default is __
    :return: the flattened dict
    """
    # http://stackoverflow.com/questions/6027558/flatten-nested-python-dictionaries-compressing-keys
    items = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, collections.MutableMapping):
            items.extend(flatten(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)


class FlatDict(dict):
    """
    A data structure to manage a flattened dict. It is initialized by passing the dict
    at time of initialization.
    """

    @property
    def dict(self):
        return self.__dict__

    def __init__(self, d):
        self.__dict__ = flatten(d)

    def __setitem__(self, key, item):
        self.__dict__[key] = item

    def __getitem__(self, key):
        return self.__dict__[key]

    def __repr__(self):
        return repr(self.__dict__)

    def __str__(self):
        return str(self.__dict__)

    def __len__(self):
        return len(self.__dict__)

    def __delitem__(self, key):
        del self.__dict__[key]

    def keys(self):
        return list(self.__dict__.keys())

    def values(self):
        return list(self.__dict__.values())

    def __cmp__(self, dictionary):
        return cmp(self.__dict__, dictionary)

    def __contains__(self, item):
        return item in self.__dict__

    def add(self, key, value):
        self.__dict__[key] = value

    def __iter__(self):
        return iter(self.__dict__)

    def __call__(self):
        return self.__dict__

    def __getattr__(self, attr):
        return self.get(attr)

'''

def main():
    d = {
        'cm_cloud': Default.cloud,
        'cm_update': '2015-06-18 22:11:48 UTC',
        'cm_user': 'gregor',
        'extra': {'created': '2015-05-21T20:37:10Z',
                  'metadata': {'base_image_ref': '398746398798372493287',
                               'description': None,
                               'image_location': 'snapshot',
                               'image_state': 'available',
                               'image_type': 'snapshot',
                               'instance_type_ephemeral_gb': '0',
                               'instance_type_flavorid': '3',
                               'instance_type_id': '1',
                               'instance_type_memory_mb': '4096',
                               'instance_type_name': 'm1.medium',
                               'instance_type_root_gb': '40',
                               'instance_type_rxtx_factor': '1.0',
                               'instance_type_swap': '0',
                               'instance_type_vcpus': '2',
                               'instance_uuid': '386473678463876387',
                               'kernel_id': None,
                               'network_allocated': 'True',
                               'owner_id': '36487264932876984723649',
                               'ramdisk_id': None,
                               'user_id': '762387463827463278649837'},
                  'minDisk': 40,
                  'minRam': 0,
                  'progress': 100,
                  'serverId': 'yiuksajhlkjahl',
                  'status': 'ACTIVE',
                  'updated': '2015-05-27T02:11:48Z'},
        'id': '39276498376478936247832687',
        'name': 'VM with Cloudmesh Configured Completely'
    }

    vm = {
        'extra': {'access_ip': '',
                  'availability_zone': 'nova',
                  'config_drive': '',
                  'created': '2015-06-19T00:06:58Z',
                  'disk_config': 'MANUAL',
                  'flavorId': '1',
                  'hostId': '',
                  'imageId': 'abcd',
                  'key': None,
                  'metadata': {},
                  'password': '********',
                  'tenantId': '1234',
                  'updated': '2015-06-19T00:06:58Z',
                  'uri': 'http://i5r.idp.iu.futuregrid.org/v2/1234/servers/abcd'},
        'id': '67f6bsf67a6b',
        'image': None,
        'name': 'gregor-cm_test',
        'private_ips': [],
        'public_ips': [],
        'size': None,
        'state': 3
    }

    pprint(d)
    pprint(flatten(d))

    f = FlatDict(d)
    pprint(f.dict)
    pprint(f.__dict__)
    print(f['cm_user'])
    print(f['extra__created'])

    print(f.cm_user)
    print(f.extra__created)

    f.cm_user = 'GREGOR'
    print(f.cm_user)

    # pprint(OpenStack_libcloud.flatten_image(d))

    # pprint(flatten(vm))

    # pprint(OpenStack_libcloud.flatten_vm(vm))


if __name__ == "__main__":
    main()

'''


from __future__ import print_function

from cloudmesh_client.shell.command import command
from cloudmesh_client.common.Printer import Printer
from cloudmesh_client.cloud.reservation import Reservation

from cloudmesh_client.shell.console import Console

import json
import pyaml

from timestring import Date

from cloudmesh_client.common.todo import TODO

from cloudmesh_client.shell.command import PluginCommand, CloudPluginCommand
from cloudmesh_client.common.Error import Error


class ReservationCommand(PluginCommand, CloudPluginCommand):
    topics = {"reservation": "todo"}

    def __init__(self, context):
        # super(self.__class__, self).__init__()
        self.context = context
        if self.context.debug:
            print("init command reservation")

    # noinspection PyUnusedLocal
    @command
    def do_reservation(self, args, arguments):
        """
        ::

            Usage:
                reservation info --user=USER --project=PROJECT
                reservation list [--name=NAME]
                                 [--user=USER]
                                 [--project=PROJECT]
                                 [--hosts=HOSTS]
                                 [--start=TIME_START]
                                 [--end=TIME_END]
                                 [--format=FORMAT]
                reservation delete [all]
                                   [--user=USER]
                                   [--project=PROJECT]
                                   [--name=NAME]
                                   [--start=TIME_START]
                                   [--end=TIME_END]
                                   [--hosts=HOSTS]
                reservation delete --file=FILE
                reservation update --name=NAME
                                  [--start=TIME_START]
                                  [--end=TIME_END]
                                  [--user=USER]
                                  [--project=PROJECT]
                                  [--hosts=HOSTS]
                                  [--description=DESCRIPTION]
                reservation add --name=NAME
                                [--start=TIME_START]
                                [--end=TIME_END]
                                [--user=USER]
                                [--project=PROJECT]
                                [--hosts=HOSTS]
                                [--description=DESCRIPTION]
                reservation add --file=FILE

            Arguments:

                NAME            Name of the reservation
                USER            Registration will be done for this user
                PROJECT         Project to be used
                HOSTS           Hosts to reserve
                TIME_START      Start time of reservation
                TIME_END        End time of reservation
                FORMAT          Format of output
                DESCRIPTION     Description for reservation
                FILE            File that contains reservation data to be added/ deleted

            Options:

                --name=NAME           Names of the reservation
                --user=USER           user name
                --project=PROJECT     project id
                --start=TIME_START    Start time of the reservation, in
                                      MM/DD/YYYY at hh:mm aa format. (default value: 01/01/1901 at 12:00 am])
                --end=TIME_END        End time of the reservation, in
                                      MM/DD/YYYY at hh:mm aa format. (default value: 12/31/2100 at 11:59 pm])
                --host=HOSTS          host name
                --description=DESCRIPTION  description summary of the reservation
                --file=FILE           Adding multiple reservations from one file
                --format=FORMAT       Format is either table, json, yaml or csv
                                      [default: table]

            Description:

                reservation info
                    lists the resources that support reservation for
                    a given user or project.
        """

        # print (arguments)
        def _print_dict(d, header=None, format='table'):
            if format == "json":
                return json.dumps(d, indent=4)
            elif format == "yaml":
                return pyaml.dump(d)
            elif format == "table":
                return Printer.write(d,
                                     order=["id",
                                            "name",
                                            "start_time",
                                            "end_time",
                                            "user",
                                            "project",
                                            "hosts",
                                            "description",
                                            "cloud"],
                                     output="table",
                                     sort_keys=True)
            elif format == "csv":
                TODO.implement()
            else:
                return d
                # return Printer.write(d,order=['cm_id, name, fingerprint'])

        def _get_db_date_format(date):
            """
            Utility Function that accepts instance of Date object and returns a string with Datetime for DB.
            :param date: Date object
            :return: Date as string with format expected in DB.
            """
            db_date = "{:}-{:}-{:} {:}:{:}".format(str(date.month).zfill(2),
                                                   str(date.day).zfill(2),
                                                   str(date.year).zfill(4),
                                                   str(date.hour).zfill(2),
                                                   str(date.minute).zfill(2))
            return db_date

        if arguments["info"]:

            TODO.implement()

        elif arguments["list"]:

            try:
                _name = arguments['--name']
                _user = arguments['--user']
                _project = arguments['--project']
                _format = arguments['--format']
                _hosts = arguments['--hosts']
                _start = arguments['--start']
                _end = arguments['--end']
                _format = arguments['--format']

                reserve = Reservation()
                dictionary = reserve.list(_name, _start, _end, _user, _project,
                                          _hosts)
                print(_print_dict(dictionary, format=_format))
                msg = "info. OK."
                Console.ok(msg)

            except Exception as e:
                Error.traceback(e)
                Console.error("Problem listing reservations")

        elif arguments["delete"]:
            if arguments["all"]:

                try:
                    reserve = Reservation()
                    reserve.delete()
                    msg = "info. OK."
                    Console.ok(msg)

                except Exception as e:
                    Error.traceback(e)
                    Console.error("Problem deleting all reservations")

            else:
                try:
                    _name = arguments['--name']
                    _user = arguments['--user']
                    _project = arguments['--project']
                    _format = arguments['--format']
                    _hosts = arguments['--hosts']
                    _start = arguments['--start']
                    _end = arguments['--end']
                    _format = arguments['--format']

                    reserve = Reservation()
                    reserve.delete(_name, _start, _end, _user, _project,
                                   _hosts)
                    msg = "info. OK."
                    Console.ok(msg)

                except Exception as e:
                    Error.traceback(e)
                    Console.error("Problem deleting reservations")

        elif arguments["add"]:

            if arguments["--file"] is None:
                name = None
                try:
                    name = arguments["--name"]
                    hosts = arguments["--hosts"]
                    user = arguments["--user"]
                    project = arguments["--project"]
                    description = arguments["--description"]

                    start_time = arguments["--start"] or "01/01/1901 at 07:30 pm"
                    end_time = arguments["--end"] or "12/31/2100 at 11:59 pm"

                    stime = Date(start_time)
                    etime = Date(end_time)

                    reserve = Reservation()
                    reserve.add(name, _get_db_date_format(stime),
                                _get_db_date_format(etime), hosts=hosts,
                                user=user,
                                project=project, description=description)

                    print("Reservation {:} added successfully".format(name))
                    msg = "info. OK."
                    Console.ok(msg)

                except Exception as e:
                    Error.traceback(e)
                    Console.error(
                        "Problem adding reservation {:}".format(name))

            else:
                try:
                    TODO.implement()
                    """
                    with open(os.path.join(sys.path[0], arguments["--file"])) as file:
                        reader = csv.reader(file)
                        for row in reader:
                            reservations = Reservation(cm_id=row[0],
                                                    label=row[1],
                                                    user=row[2],
                                                    project=row[3],
                                                    start_time=row[4],
                                                    end_time=row[5],
                                                    host=row[6],
                                                    summary=row[7])
                            db.add()
                    """
                except Exception as e:
                    print("Error in adding from file. ", e)

        elif arguments["update"]:
            name = None
            try:
                name = arguments["--name"]
                hosts = arguments["--hosts"]
                user = arguments["--user"]
                project = arguments["--project"]
                description = arguments["--description"]

                start_time = arguments["--start"] or "01/01/1901 at 07:30 pm"
                end_time = arguments["--end"] or "12/31/2100 at 11:59 pm"

                stime = Date(start_time)
                etime = Date(end_time)

                reserve = Reservation()
                reserve.update(name,
                               _get_db_date_format(stime),
                               _get_db_date_format(etime),
                               hosts=hosts,
                               user=user,
                               project=project,
                               description=description)

                print("Reservation {:} updated successfully".format(name))
                msg = "info. OK."
                Console.ok(msg)

            except Exception as e:
                Error.traceback(e)
                Console.error("Problem updating reservation {:}".format(name))

        return ""



from __future__ import print_function
from cloudmesh_client.shell.command import command
from cloudmesh_client.shell.command import PluginCommand, CloudPluginCommand

from cloudmesh_client.common.ConfigDict import ConfigDict
from cloudmesh_client.setup import os_execute


class PortalCommand(PluginCommand, CloudPluginCommand):
    topics = {"portal": "cloud"}

    def __init__(self, context):
        self.context = context
        if self.context.debug:
            print("init command portal")

    def get_conf(self):
        config = ConfigDict("cloudmesh.yaml")
        data = dict(config["cloudmesh.portal"])
        print(data)
        return data

    # noinspection PyUnusedLocal
    @command
    def do_portal(self, args, arguments):
        """
        ::

            Usage:
                portal start
                portal stop

            Examples:
                portal start
                    starts the portal and opens the default web page

                portal stop
                    stops the portal

        """
        if arguments["start"]:
            ValueError("Not yet implemented")

            data = self.get_conf()
            commands = """
                cd {location}; make run &
                cd {location}; make view &
                """.format(**data)
            print(data)
            os_execute(commands)

        if arguments["start"]:
            ValueError("Not yet implemented")

            data = self.get_conf()
            commands = """
                cd {location}; make run
                """.format(**data)
            print(data)
            os_execute(commands)

        elif arguments["stop"]:
            ValueError("Not yet implemented")

        return ""




from __future__ import print_function
from cloudmesh_client.default import Default
from cloudmesh_client.shell.command import command
from cloudmesh_client.shell.console import Console
from cloudmesh_client.cloud.usage import Usage
from cloudmesh_client.shell.command import PluginCommand, CloudPluginCommand


class UsageCommand(PluginCommand, CloudPluginCommand):
    topics = {"usage": "cloud"}

    def __init__(self, context):
        self.context = context
        if self.context.debug:
            print("init command list")

    # noinspection PyUnusedLocal
    @command
    def do_usage(self, args, arguments):
        """
        ::

            Usage:
                usage list [--cloud=CLOUD] [--start=START] [--end=END] [--tenant=TENANT] [--format=FORMAT]

                Show usage data.

            Options:
               --format=FORMAT  the output format [default: table]
               --cloud=CLOUD    the cloud name
               --tenant=TENANT  the tenant name
               --start=START    Usage range start date ex 2012-01-20, default is: 4 weeks ago
               --end=END        Usage range end date, ex 2012-01-20, default is: tomorrow


            Examples:
                cm usage list

        """

        if arguments["list"]:
            cloud = arguments["--cloud"] or Default.get(name="cloud")

            if not cloud:
                Console.error("cloud doesn't exist")
                return ""
            output_format = arguments["--format"]
            start = arguments["--start"]
            end = arguments["--end"]
            tenant = arguments["--tenant"]
            usage = Usage.list(cloud,
                               start=start,
                               end=end,
                               tenant=tenant,
                               format=output_format)
            Console.msg(usage)
            return ""



class DictDB(dict):
    """

    Implements a persistent dictionary using sqlite

    Usage:

        element = {"name":, "abc",
                    "address": "here",
                    "email": "abc@example.com"}

        The name in the dict is identifies where to store it.

        d = PersistentData(os.path.join("~", ".cloudmesh/cloudmesh.db"))

        d.add(element)

        d.delete("name": "abc")
            # deletes all elements from the db where ethe name is matching abc

        d.json
            # exports to json

        d.yaml
            # exports a yaml as a string

        d.get("name": "abc")
            gets the first match

        d.find("name": "abc")
            finds the first match


    """

    def __init__(self, filename, d=None, attributes=None):
        """
        initializes the dict database

        :param filename: filename where to store the database. The name can include ~.
        :param d: an optional dict that is provided for initialization. If the elements
                  are already in the database they will be overwritten with the new dict.
                  Typically we initialize without a dict.
        :param attributes: An optional list of attributes to be passed
        """
        pass

    def attributes(self, attribute_names):
        """
        defines the attributes that each record can have. If a database has previously
        defined and new attributes are added the database records are updated.

        :param attribute_names: The attribute names
        """
        pass

    def add(self, d):
        """
        adds the dict to the database. The attribute "name" is used to define a
        unique name for the object in the database

        :param d: the dict that at minimum must contain a name attribute
        """
        pass

    def delete(self, operator="and", **kwargs):
        """
        deletes all elements in that match the query formulated by kwargs and the operator
        operators that are allowed are = and, or

        :param kwargs: the attributes that we look for
        :param operator: the operator and / or
        """
        pass

    def find(self, operator="and", **kwargs):
        """
        finds all elements in that match the query formulated by kwargs and the operator
        operators that are allowed are = and, or

        :param kwargs: the attributes that we look for
        :param operator: the operator and / or
        """
        pass

    def get(self, operator="and", **kwargs):
        """
        finds the first elements in that match the query formulated by kwargs and the operator
        operators that are allowed are = and, or

        :param kwargs: the attributes that we look for
        :param operator: the operator and / or
        """
        pass

    @property
    def json(self):
        """
        :return: the json object of the database
        """
        return None

    @property
    def yaml(self):
        """
        :return: a yaml string of the database
        """
        return None

    def backup(self, filename=None):
        """
        backs up the current database. If the filename is omitted, the backup will be
        created in the same directory as the database with the postfix .bak.#
        where # is the largest number that has not yet been assigned. For example.,
        lets assume a backup exists with the name file.bak.9, than the next higher
        number is used (e.g. 10) and the backup file file.back.10 is used.

        :param filename: the backup filename.
        :return:
        """
        pass



from __future__ import print_function
from cloudmesh_client.common.ConfigDict import ConfigDict

from cloudmesh_client.common.todo import TODO
# add imports for other cloud providers in future
from cloudmesh_client.shell.console import Console
from cloudmesh_client.cloud.ListResource import ListResource
from cloudmesh_client.common.Printer import Printer
from cloudmesh_client.db import CloudmeshDatabase
from cloudmesh_client.cloud.iaas.CloudProvider import CloudProvider
from cloudmesh_client.common.Error import Error
from uuid import UUID
from cloudmesh_client.common.dotdict import dotdict
from builtins import input
from pprint import pprint
from cloudmesh_client.common.ConfigDict import Config
from cloudmesh_client.default import Default
from cloudmesh_client.common.menu import menu_return_num
from cloudmesh_client.common.SSHkey import SSHkey
import os
from os.path import expanduser
import requests
import subprocess

# noinspection PyPep8Naming
class Key(ListResource):
    cm = CloudmeshDatabase()

    @classmethod
    def info(cls, **kwargs):
        raise NotImplementedError()

    @classmethod
    def get_from_dir(cls, directory=None, store=True):
        directory = directory or Config.path_expand("~/.ssh")
        files = [file for file in os.listdir(expanduser(Config.path_expand(directory)))
                 if file.lower().endswith(".pub")]
        d = []
        for file in files:
            location = Config.path_expand("{:}/{:}".format(directory, file))

            sshkey = SSHkey(location).get()

            i = sshkey["comment"]
            if i is not None:
                i = i.replace("@", "_")
                i = i.replace("-", "_")
                i = i.replace(" ", "_")
                i = i.replace(".", "_")
            else:
                # use base name
                i = file.replace(".pub", "")
            sshkey["kind"] = "key"
            sshkey["source"] = 'file'

            if store:
                cls._add_from_sshkey(
                    dict(sshkey),
                    keyname=sshkey["name"],
                    source=sshkey["source"],
                    uri=sshkey["uri"])
            else:
                d.append(dict(sshkey))
        if not store:
            return d

    @classmethod
    def get_from_git(cls, username, store=True):
        """

        :param username: the github username
        :return: an array of public keys
        :rtype: list
        """
        uri = 'https://github.com/{:}.keys'.format(username)
        content = requests.get(uri).text.strip("\n").split("\n")

        d = []

        for key in range(0, len(content)):
            value = content[key]
            thekey = {}

            name = "{}_git_{}".format(username, key)

            thekey = {
                'uri': uri,
                'string': value,
                'fingerprint': SSHkey._fingerprint(value),
                'name': name,
                'comment': name,
                'cm_id': name,
                'source': 'git',
                'kind': 'key'
            }

            thekey["type"], thekey["key"], thekey["comment"] = SSHkey._parse(value)

            if thekey["comment"] is None:
                thekey["comment"] = name
            d.append(thekey)
            if store:
                try:

                    cls.cm.add(thekey)
                except:
                    Console.error("Key already in db", traceflag=False)
        if not store:
            return d
                # noinspection PyProtectedMember,PyUnreachableCode,PyUnusedLocal

    @classmethod
    def get_from_yaml(cls, filename=None, load_order=None, store=True):
        """
        :param filename: name of the yaml file
        :return: a SSHKeyManager (dict of keys)
        """
        config = None
        if filename is None:
            # default = Config.path_expand(os.path.join("~", ".cloudmesh", "cloudmesh.yaml"))
            # config = ConfigDict("cloudmesh.yaml")
            filename = "cloudmesh.yaml"
            config = ConfigDict(filename)
        elif load_order:
            config = ConfigDict(filename, load_order)
        else:
            Console.error("Wrong arguments")
            return
        config_keys = config["cloudmesh"]["keys"]
        default = config_keys["default"]
        keylist = config_keys["keylist"]

        uri = Config.path_expand(os.path.join("~", ".cloudmesh", filename))


        d = []
        for key in list(keylist.keys()):
            keyname = key
            value = keylist[key]
            if os.path.isfile(Config.path_expand(value)):
                path = Config.path_expand(value)
                if store:
                    Key.add_from_path(path, keyname)
                else:
                    d.append(Key.add_from_path(path, keyname, store=False))
            else:

                keytype, string, comment = SSHkey._parse(value)
                thekey = {
                    'uri': 'yaml://{}'.format(uri),
                    'string': value,
                    'fingerprint': SSHkey._fingerprint(value),
                    'name': keyname,
                    'comment': comment,
                    'source': 'git',
                    'kind': 'key'
                }

                thekey["type"], thekey["key"], thekey["comment"] = SSHkey._parse(value)

                if thekey["comment"] is None:
                    thekey["comment"] = keyname
                if store:
                    try:
                        cls.cm.add(thekey)
                    except:
                        Console.error("Key already in db", traceflag=False)
                else:
                    d.append(thekey)
        if not store:
            return d



        """
        take a look into original cloudmesh code, its possible to either specify a key or a filename
        the original one is able to figure this out and do the rightthing. We may want to add this
        logic to the SSHkey class, so we can initialize either via filename or key string.
        It would than figure out the right thing

        cloudmesh:
          keys:
            idrsa: ~/.ssh/id_rsa.pub

        cloudmesh:
        ...
          keys:
            default: name of the key
            keylist:
              keyname: ~/.ssh/id_rsa.pub
              keyname: ssh rsa hajfhjldahlfjhdlsak ..... comment
              github-x: github
        """
    @classmethod
    def get_from_cloud(cls, cloud, live=False, format="table"):
        """
        This method lists all keys of the cloud
        :param cloud: the cloud name
        :return: a SSHKeyManager (dict of keys)
        """
        try:
            return CloudProvider(cloud).provider.list_key(cloud)

        except Exception as ex:
            Console.error(ex.message)

    @classmethod
    def _add_from_sshkey(cls,
                         sshkey,
                         keyname=None,
                         user=None,
                         source=None,
                         uri=None):

        user = user or cls.cm.user

        if keyname is None:
            try:
                keyname = sshkey['name']
            except:
                pass
        if keyname is None:
            print("ERROR: keyname is None")

        thekey = {
            "kind": "key",
            "name": keyname,
            "uri": sshkey['uri'],
            "source": sshkey['source'],
            "fingerprint": sshkey['fingerprint'],
            "comment": sshkey['comment'],
            "value": sshkey['string'],
            "category": "general",
            "user": user}

        cls.cm.add(thekey)

    @classmethod
    def add_key_to_cloud(cls, user, keyname, cloud):
        """

        :param user:
        :param keyname:
        :param cloud:
        :param name_on_cloud:
        """

        key = cls.cm.find(kind="key", name=keyname, scope="first")
        if key is None:
            Console.error("Key with the name {:} not found in database.".format(keyname))
            return

        try:
            if cloud is not None:
                print("Adding key {:} to cloud {:}".format(keyname, cloud))
                cloud_provider = CloudProvider(cloud).provider
                cloud_provider.add_key_to_cloud(keyname, key["value"])
        except Exception as e:
            Console.error("problem uploading key {} to cloud {}: {}".format(keyname, cloud, e.message), traceflag=False)

    @classmethod
    def list(cls, category=None, live=False, output="table"):
        "this does not work only returns all ceys in the db"
        (order, header) = CloudProvider(category).get_attributes("key")
        d = cls.cm.find(kind="key", scope="all", output=output)
        return Printer.write(d,
                             order=order,
                             header=header,
                             output=output)

    @classmethod
    def list_on_cloud(cls, cloud, live=False, format="table"):
        """
        This method lists all keys of the cloud
        :param cloud: the cloud name
        """
        try:
            keys = CloudProvider(cloud).provider.list_key(cloud)
            for key in keys:
                keys[key]["category"] = cloud
            if keys is None or keys is []:
                return None

            (order, header) = CloudProvider(cloud).get_attributes("key")

            return Printer.write(keys,
                                 order=order,
                                 header=header,
                                 output=format)
        except Exception as ex:
            Console.error(ex.message)

    @classmethod
    def run_command(cls, cmd):
        """ Runs a command in a shell, returns the result"""
        p = subprocess.Popen(cmd, shell=True,
                         stdout=subprocess.PIPE)
        return p.stdout.read()

    @classmethod
    def add_azure_key_to_db(cls, key_name, key_path, certificate_path, pfx_path):
        """
            Adds the public key to the existing database model and adds the certificate, key and
            fingerprint into the azure key database model.
        :param key_name: Key name to be added
        :param key_path: Public key path
        :param certificate_path: Certificate file path(PEM file)
        :param pfx_path: PKCS encoded certificate path
        :return:
        """
        pprint("add_azure_key_to_db")
        # Add to the current DB
        cls.add_from_path(key_path,
                            key_name,
                            source="ssh",
                            uri="file://" + key_path)

        # Add certificate to the new DB
        fingerprint_cmd = "openssl x509 -in "+certificate_path+" -sha1 -noout -fingerprint | sed s/://g"
        # print("fingerprint_cmd:", fingerprint_cmd)
        fingerprint = cls.run_command(fingerprint_cmd)
        fingerprint = fingerprint.split('=')[1]
        fingerprint = fingerprint.rstrip('\n')
        # pprint("Certificate Fingerprint="+fingerprint)
        key_azure_obj = {
            "kind": "key_azure",
            "name": key_name,
            "fingerprint": fingerprint,
            "certificate": certificate_path,
            "key_path": key_path,
            "pfx_path": pfx_path}
        cls.cm.add(key_azure_obj)
        Console.info("Azure key added.ok.")

    @classmethod
    def clear(cls, **kwargs):
        raise NotImplementedError()

    @classmethod
    def refresh(cls, **kwargs):
        raise NotImplementedError()

    @classmethod
    def delete(cls, name=None, cloud=None):

        if cloud is not None and name is not None:
            result = CloudProvider(cloud).provider.delete_key_from_cloud(name)
        elif cloud is not None and name is None:
            #
            # get a list of keys from cloud
            # loop over all keys and use the provider delete from cloud to delete that key

            Console.error("delete all keys from cloud not yet implemented")

        if name is None:
            cls.cm.delete(kind="key", provider="general")
        else:
            cls.cm.delete(name=name, kind="key", provider="general")

    @classmethod
    def all(cls, output="dict"):
        return cls.cm.find(kind="key", scope="all", output=output)


    @classmethod
    def find(cls, name=None, output="dict"):
        return cls.get(name=name, output=output)

    @classmethod
    def get(cls, name=None, output="dict"):
        """
        Finds the key on the database by name

        :param name: name of the key to be found
        :return: Query object of the search
        """
        if name is None:
            return cls.cm.find(kind="key", output=output)
        else:
            return cls.cm.find(kind="key", name=name, output=output, scope="first")

    @classmethod
    def set_default(cls, name):
        Default.set_key(name)

    # deprecated use Default.key
    @classmethod
    def get_default(cls):
        return Default.key

    @classmethod
    def delete_from_cloud(cls, name, cloud=None):
        pass

    @classmethod
    def _delete_from_db(cls, name=None):
        if name is None:
            cls.cm.delete(kind='key')
        else:
            cls.cm.delete(kind='key', name=name)

    @classmethod
    def select(cls):
        options = []
        d = cls.get(output='dict')
        for i in d:
            line = '{}: {}'.format(d[i]['name'], d[i]['fingerprint'])
            options.append(line)
        num = menu_return_num('KEYS', options)
        if num != 'q':
            return options[num]
        return num

    #
    # ADD
    #
    @classmethod
    def add_from_path(cls,
                      path,
                      keyname=None,
                      user=None,
                      source=None,
                      uri=None,
                      store=True):
        """
        Adds the key to the database based on the path

        :param keyname: name of the key or path to the key
        :return:
        """

        user = user or cls.cm.user

        sshkey = SSHkey(Config.path_expand(path))

        if store:
            cls._add_from_sshkey(sshkey.__key__,
                             keyname,
                             user,
                             source=source,
                             uri=uri)
        else:
            return sshkey.__key__



from future.utils import iteritems
import json
from datetime import datetime
import textwrap

from cloudmesh_client.common.Shell import Shell
from cloudmesh_client.common.Printer import Printer
from cloudmesh_client.common.TableParser import TableParser
from cloudmesh_client.common.ConfigDict import Config, ConfigDict
from cloudmesh_client.cloud.hpc.BatchProviderBase import BatchProviderBase
from cloudmesh_client.db import CloudmeshDatabase
import os
from cloudmesh_client.common.Error import Error


# noinspection PyBroadException
class BatchProviderSLURM(BatchProviderBase):
    cm = CloudmeshDatabase()

    kind = "slurm"

    @classmethod
    def queue(cls, cluster, format='json', job=None):
        try:
            args = 'squeue '
            if job is not None:
                if job.isdigit():
                    args += ' -j {} '.format(str(job))  # search by job id
                else:
                    args += ' -n {} '.format(job)  # search by job name
            f = '--format=%all'
            args += f
            result = Shell.ssh(cluster, args)

            # TODO: process till header is found...(Need a better way)
            l = result.splitlines()
            for i, res in enumerate(l):
                if 'ACCOUNT|GRES|' in res:
                    result = "\n".join(str(x) for x in l[i:])
                    break

            parser = TableParser(strip=True)
            d = parser.to_dict(result)

            # add cluster and updated to each entry
            for key in list(d.keys()):
                d[key]['cluster'] = cluster
                d[key]['updated'] = datetime.now().strftime("%Y-%m-%d %H:%M:%S")

            if format == 'json':
                return json.dumps(d, indent=4, separators=(',', ': '))

            else:
                return (Printer.write(d,
                                      order=['cluster',
                                             'jobid',
                                             'partition',
                                             'name',
                                             'user',
                                             'st',
                                             'time',
                                             'nodes',
                                             'nodelist',
                                             'updated'],
                                      output=format))
        except Exception as e:
            Error.traceback(e)
            return e

    @classmethod
    def info(cls, cluster, format='json', all=False):

        if all:
            result = Shell.ssh(cluster, 'sinfo --format=\"%all\"')
        else:
            result = Shell.ssh(
                cluster,
                'sinfo --format=\"%P|%a|%l|%D|%t|%N\"')

        # ignore leading lines till header is found
        l = result.splitlines()
        for i, res in enumerate(l):
            if 'PARTITION|AVAIL|' in res:
                result = "\n".join(l[i:])
                break

        parser = TableParser(strip=False)
        d = parser.to_dict(result)

        # add cluster and updated to each entry
        for key in list(d.keys()):
            d[key]['cluster'] = cluster
            d[key]['updated'] = datetime.now().strftime("%Y-%m-%d %H:%M:%S")

        if format == 'json':
            return json.dumps(d, indent=4, separators=(',', ': '))

        else:
            return (Printer.write(d,
                                  order=['cluster',
                                         'partition',
                                         'avail',
                                         'timelimit',
                                         'nodes',
                                         'state',
                                         'nodelist',
                                         'updated'],
                                  output=format))

    @classmethod
    def test(cls, cluster, time):
        result = Shell.ssh(cluster,
                           "srun -n1 -t {} echo '#CLOUDMESH: Test ok'".format(
                               time))
        return result

    @classmethod
    def run(cls, cluster, group, cmd, **kwargs):

        # determine the script name..

        #
        # TODO: script count is variable in data base, we test if fil exists and if it
        # does increase counter till we find one that does not, that will be new counter.
        # new counter will than be placed in db.
        #
        # define get_script_name(directory, prefix, counter)
        # there maybe s a similar thing already in the old cloudmesh
        #

        # if not kwargs['-name']:
        #
        #    old_count = Shell.ssh(cluster,
        #                          "ls {}*.sh | wc -l | sed 's/$/ count/'".
        #                          format(username))
        #    c = [f for f in old_count.splitlines() if 'count' in f]
        #    script_count = c[0].split()[0]
        # else:
        #    script_count = kwargs['-name']

        config = cls.read_config(cluster)
        if config["credentials"]["username"] == 'TBD':
            return "Please enter username in cloudmesh.yaml for cluster {}".format(cluster)

        cls.incr()
        data = {
            "cluster": cluster,
            "count": cls.counter(),
            "username": config["credentials"]["username"],
            "remote_experiment_dir": config["default"]["experiment_dir"],
            "queue": config["default"]["queue"],
            "id": None,
            "nodes": 1,
            "tasks_per_node": 1,
        }
        data["script_base_name"] = "{username}-{count}".format(**data)
        data["script_name"] = "{username}-{count}.sh".format(**data)
        data["script_output"] = "{username}-{count}.out".format(**data)
        data["script_error"] = "{username}-{count}.err".format(**data)
        data["remote_experiment_dir"] = \
            "{remote_experiment_dir}/{count}".format(**data).format(**data)
        data["group"] = group

        # overwrite defaults
        option_mapping = {'-t': '{tasks_per_node}'.format(**data),
                          '-N': '{nodes}'.format(**data),
                          '-p': '{queue}'.format(**data),
                          '-o': '{script_output}'.format(**data),
                          '-D': '{remote_experiment_dir}'.format(**data),
                          '-e': '{script_error}'.format(**data)}

        # map(lambda k, v:
        #    option_mapping.__setitem__(k, kwargs.get(k) or v),
        #    option_mapping.items())
        #
        # rewrite for better readability
        for (k, v) in iteritems(option_mapping):
            option_mapping[k] = kwargs.get(k) or v

        config = cls.read_config(cluster)
        project = None
        try:
            project = config["credentials"]["project"]
            if project.lower() not in ["tbd", "none"]:
                option_mapping["-A"] = project
        except:
            pass

        for key in option_mapping:
            data[key] = option_mapping[key]

        # create the options for the script
        options = ""
        for key, value in option_mapping.items():
            options += '#SBATCH {} {}\n'.format(key, value)

        cls.create_remote_dir(cluster, data["remote_experiment_dir"])

        # if the command is a script, copy the script
        if os.path.isfile(Config.path_expand(cmd)):
            _from = Config.path_expand(cmd)
            _to = '{cluster}:{remote_experiment_dir}'.format(**data)

            local_file_name = cmd.split('/')[-1]
            Shell.execute("rsync", [_from, _to])
            data["command"] = '{remote_experiment_dir}/{local_file_name}'.format(local_file_name=local_file_name,
                                                                                 **data)
        else:
            data["command"] = cmd

        data["options"] = options

        script = textwrap.dedent(
            """
            #! /bin/sh
            {options}

            echo '#CLOUDMESH: BATCH ENVIRONMENT'
            echo 'BASIL_RESERVATION_ID:' $BASIL_RESERVATION_ID
            echo 'SLURM_CPU_BIND:' $SLURM_CPU_BIND
            echo 'SLURM_JOB_ID:' $SLURM_JOB_ID
            echo 'SLURM_JOB_CPUS_PER_NODE:' $SLURM_JOB_CPUS_PER_NODE
            echo 'SLURM_JOB_DEPENDENCY:' $SLURM_JOB_DEPENDENCY
            echo 'SLURM_JOB_NAME:' $SLURM_JOB_NAME
            echo 'SLURM_JOB_NODELIST:' $SLURM_JOB_NODELIST
            echo 'SLURM_JOB_NUM_NODES:' $SLURM_JOB_NUM_NODES
            echo 'SLURM_MEM_BIND:' $SLURM_MEM_BIND
            echo 'SLURM_TASKS_PER_NODE:' $SLURM_TASKS_PER_NODE
            echo 'MPIRUN_NOALLOCATE:' $MPIRUN_NOALLOCATE
            echo 'MPIRUN_NOFREE:' $MPIRUN_NOFREE
            echo 'SLURM_NTASKS_PER_CORE:' $SLURM_NTASKS_PER_CORE
            echo 'SLURM_NTASKS_PER_NODE:' $SLURM_NTASKS_PER_NODE
            echo 'SLURM_NTASKS_PER_SOCKET:' $SLURM_NTASKS_PER_SOCKET
            echo 'SLURM_RESTART_COUNT:' $SLURM_RESTART_COUNT
            echo 'SLURM_SUBMIT_DIR:' $SLURM_SUBMIT_DIR
            echo 'MPIRUN_PARTITION:' $MPIRUN_PARTITION
            d=$(date)
            echo \"#CLOUDMESH: status, start, $d\"
            srun -l echo \"#CLOUDMESH: status, start, $d\"
            srun -l {command}
            d=$(date)
            srun -l echo \"#CLOUDMESH: status, finished, $d\"
            d=$(date)
            echo \"#CLOUDMESH: status, finished, $d\"
            """
        ).format(**data).replace("\r\n", "\n").strip()

        _from = Config.path_expand('~/.cloudmesh/{script_name}'.format(**data))
        _to = '{cluster}:{remote_experiment_dir}'.format(**data)
        data["from"] = _from
        data["to"] = _to
        data["script"] = script
        # write the script to local

        # print(_from)
        # print(_to)

        with open(_from, 'w') as local_file:
            local_file.write(script)

        # copy to remote host
        Shell.scp(_from, _to)

        # delete local file
        # Shell.execute('rm', _from)
        # import sys; sys.exit()

        # run the sbatch command

        cmd = 'sbatch {remote_experiment_dir}/{script_name}'.format(**data)
        data["cmd"] = cmd

        # print ("CMD>", cmd)
        result = Shell.ssh(cluster, cmd)

        data["output"] = result

        # find id
        for line in result.split("\n"):
            # print ("LLL>", line)
            if "Submitted batch job" in line:
                data["job_id"] = int(line.replace("Submitted batch job ", "").strip())
                break

        #
        # HACK, should not depend on Model.py
        #

        # from cloudmesh_client.db.model import BATCHJOB
        # name = ""
        # BATCHJOB(name,
        #         cluster=data["cluster"],
        #         id=data["id"],
        #         script=data["script"]) # has user and username which seems wrong

        # here what we have in data and want to store the - options are obviously wrong
        # and need to be full names
        # noinspection PyPep8,PyPep8
        """
                {'-D': '/N/u/gvonlasz/experiment/3',
                 '-N': '1',
                 '-o': 'gvonlasz-3.out',
                 '-p': 'delta',
                 '-t': '1',
                 'cluster': 'india',
                 'cmd': 'sbatch /N/u/gvonlasz/experiment/3/gvonlasz-3.sh',
                 'command': 'uname',
                 'count': 3,
                 'from': '/Users/big/.cloudmesh/gvonlasz-3.sh',
                 'id': 1346,
                 'options': '#SBATCH -t 1\n#SBATCH -o gvonlasz-3.out\n#SBATCH -N 1\n#SBATCH -p delta\n#SBATCH -D /N/u/gvonlasz/experiment/3\n',
                 'output': 'Submitted batch job 1346',
                 'queue': 'delta',
                 'remote_experiment_dir': '/N/u/gvonlasz/experiment/3',
                 'script': "#! /bin/sh\n#SBATCH -t 1\n#SBATCH -o gvonlasz-3.out\n#SBATCH -N 1\n#SBATCH -p delta\n#SBATCH -D /N/u/gvonlasz/experiment/3\n\nsrun -l echo '#CLOUDMESH: Starting'\nsrun -l uname\nsrun -l echo '#CLOUDMESH: Test ok'",
                 'script_base_name': 'gvonlasz-3',
                 'script_name': 'gvonlasz-3.sh',
                 'script_output': 'gvonlasz-3.out',
                 'to': 'india:/N/u/gvonlasz/experiment/3',
                 'username': 'gvonlasz'}
                """

        """
        we also want to store what part of the .out file,

        BASIL_RESERVATION_ID:
        SLURM_CPU_BIND:
        SLURM_JOB_ID: 1351
        SLURM_JOB_CPUS_PER_NODE: 12
        SLURM_JOB_DEPENDENCY:
        SLURM_JOB_NAME: gvonlasz-8.sh
        SLURM_JOB_NODELIST: d001
        SLURM_JOB_NUM_NODES: 1
        SLURM_MEM_BIND:
        SLURM_TASKS_PER_NODE: 12
        MPIRUN_NOALLOCATE:
        MPIRUN_NOFREE:
        SLURM_NTASKS_PER_CORE:
        SLURM_NTASKS_PER_NODE:
        SLURM_NTASKS_PER_SOCKET:
        SLURM_RESTART_COUNT:
        SLURM_SUBMIT_DIR: /N/u/gvonlasz
        MPIRUN_PARTITION:

        so maybe we want to use some of the names here as they reflect the env vars
        """

        #
        # add data to database
        #
        # remove the - options

        for key in ['-t', '-N', '-p', '-o', '-D', '-e']:
            if key in data:
                print(key, data[key])
                del data[key]
        data['status'] = 'started'
        cls.add_db(**data)

        return data

    @classmethod
    def delete(cls, cluster, job, group=None):
        """
        This method is used to terminate a job with the specified or a group of jobs
        job_id or job_name in a given cluster
        :param group:
        :param cluster: the cluster like comet
        :param job: the job id or name
        :return: success message or error
        """
        try:
            if group is not None:
                # get the job ids from the db

                arguments = {'cluster': cluster,
                             'group': group}
                db_jobs = cls.cm.find('batchjob',
                                      **arguments)

                list1 = []
                for i in db_jobs:
                    list1.append(db_jobs[i]['job_id'])

                # read active jobs
                active_jobs = json.loads(cls.queue(cluster))
                list2 = []
                for i in active_jobs:
                    list2.append(active_jobs[i]['jobid'])

                # find intersection
                res = set(list1).intersection(set(list2))

                if res is not None:
                    for j in res:
                        cmd = 'scancel {}'.format(str(j))
                        Shell.ssh(cluster, cmd)
                        print("Deleted {}".format(j))

                return "All jobs for group {} killed successfully".format(group)

            else:
                args = 'scancel '
                if job.isdigit():
                    args += job
                else:
                    args += "-n {}".format(job)

                Shell.ssh(cluster, args)
                return "Job {} killed successfully".format(job)
        except Exception as ex:
            print("in exceptio")
            print(ex)
            return ex

    @classmethod
    def add_db(cls, **kwargs):

        kwargs['name'] = kwargs.get('script_name')

        db_obj = {0: {"batchjob": kwargs}}
        cls.cm.add_obj(db_obj)
        cls.cm.save()



from cloudmesh_client.common.ConfigDict import ConfigDict
from cloudmesh_client.cloud.iaas.provider.openstack.CloudProviderOpenstackAPI import \
    CloudProviderOpenstackAPI
from cloudmesh_client.cloud.iaas.CloudProviderBase import CloudProviderBase
import requests
from cloudmesh_client.common.Error import Error
from cloudmesh_client.common.todo import TODO
from cloudmesh_client.cloud.iaas.provider.libcloud.CloudProviderLibcloudEC2 import CloudProviderLibcloudEC2
from cloudmesh_client.cloud.iaas.provider.azure.CloudProviderAzureAPI import CloudProviderAzureAPI
from cloudmesh_client.shell.console import Console

requests.packages.urllib3.disable_warnings()


class CloudProvider(CloudProviderBase):
    def __init__(self, cloudname, user=None, flat=True):
        super(CloudProvider, self).__init__(cloudname, user=user)

        try:
            d = ConfigDict("cloudmesh.yaml")

            if cloudname not in d["cloudmesh"]["clouds"]:
                Console.error("the cloud {} is not defined in the yaml file. failed."
                                 .format(cloudname), traceflag=False)
                return

            cloud_details = d["cloudmesh"]["clouds"][cloudname]

            if cloud_details["cm_type"] == "openstack":
                provider = CloudProviderOpenstackAPI(
                    cloudname,
                    cloud_details,
                    flat=flat)
                self.provider = provider
                self.provider_class = CloudProviderOpenstackAPI

            if cloud_details["cm_type"] == "ec2":
                provider = CloudProviderLibcloudEC2(
                    cloudname,
                    cloud_details,
                    flat=flat)
                self.provider = provider
                self.provider_class = CloudProviderLibcloudEC2

            if cloud_details["cm_type"] == "azure":
                # raise ValueError("azure cloud provider yet implemented. failed.")
                # Console.TODO("Azure provider to be implemented")
                provider = CloudProviderAzureAPI(
                    cloudname,
                    cloud_details)
                self.provider = provider
                self.provider_class = CloudProviderAzureAPI



        except Exception as e:
            Error.traceback(e)

    def get_attributes(self, kind):
        return self.provider.attributes(kind)


def main():
    from pprint import pprint

    cloud = "kilo"
    provider = CloudProvider(cloud).provider

    print(provider, type(provider))

    # pprint (provider.__dict__)
    # pprint (dir(provider))

    r = provider.list_flavor(cloud)
    pprint(r)

    for kind in ["flavor", "image", "vm", "key"]:  # "limits", "quota"]:
        r = provider.list(kind, cloud)
        pprint(r)


if __name__ == "__main__":
    main()



from __future__ import print_function

import os.path
import sys

import hostlist
import yaml

from cloudmesh_client.cloud.ListResource import ListResource
from cloudmesh_client.common.ConfigDict import ConfigDict
from cloudmesh_client.common.Printer import Printer
from cloudmesh_client.common.Shell import Shell
from cloudmesh_client.common.util import banner
from cloudmesh_client.shell.console import Console


class Inventory(ListResource):
    def info(self):
        banner("Configuration")
        Console.ok('Object Attibutes: {:}'.format(', '.join(self.order)))
        Console.ok('Objects: {:}'.format(len(self.data)))
        Console.ok('Configuration File: {:}'.format(self.config_filename))
        Console.ok('Configuration:')

        print(self.config)

        try:
            config = ConfigDict(filename=self.config_filename)
        except Exception as e:
            Console.error("Problem reading the yaml file {:}".format(
                self.config_filename), traceflag=False)
            Console.error("Please check if the file exists or is empty", traceflag=False)
            Console.error(e.message)

        banner("")

    def __init__(self):

        self.order = [
            "host",
            "cluster",
            "label",
            "service",
            "ip",
            "project",
            "owners",
            "comment"]

        self.entry = {}
        for key in self.order:
            self.entry[key] = ""

        self.data = {}

        self.config_filename = "cloudmesh.yaml"
        self.config = ConfigDict(filename=self.config_filename)

        self.datafile = self.config["cloudmesh.system.data"]
        self.read(self.datafile)

    def read(self, filename=None, format="yaml"):
        if filename is None:
            filename = self.datafile

        if not os.path.isfile(filename):
            self.save(filename)
        stream = open(filename, "r")
        self.data = yaml.safe_load(stream)
        stream.close()

    def save(self, filename=None, format="yaml"):
        if filename is None:
            filename = self.datafile
        with open(filename, 'w') as yaml_file:
            yaml_file.write(self.list(format=format))

    def add(self, **kwargs):

        if "host" not in kwargs:
            print("ERROR no id specified")
            sys.exit(1)

        hosts = hostlist.expand_hostlist(kwargs['host'])

        for host in hosts:
            if host in self.data:
                entry = self.data[host]
            else:
                entry = dict(self.entry)
                self.data[host] = entry
            for key, value in kwargs.items():
                entry[key] = value
            entry['host'] = host
            for attribute in entry:
                self.data[host][attribute] = entry[attribute]
        self.save()

    def list(self, format='dict', sort_keys=True, order=None):
        if order is None:
            order = self.order
        return Printer.write(self.data,
                             order=order,
                             output=format,
                             sort_keys=sort_keys)

    def _str(self, data, with_empty=False):
        print
        for key in data:
            if self.data[key] is '' or self.data[key] is None:
                pass
            else:
                print(self.data[key])


# noinspection PyBroadException,PyPep8Naming
class command_system(object):
    @classmethod
    def status(cls, host):
        msg = "Unknown host"
        try:
            msg = Shell.ping("-c", "1", host)
        except:
            pass
        if "1 packets transmitted, 1 packets received" in msg:
            return True
        elif "Unknown host" in msg:
            return False
        else:
            return False


if __name__ == "__main__":
    i = Inventory()
    banner("Info")
    i.info()

    banner("changing values")
    i.add(host="i1", cluster="kilo", label="india")
    i.add(host="i2", cluster="kilo", label="gregor")
    i.add(host="d[1-4]", cluster="kilo", label="delta")

    banner("saving")
    i.save()

    for output in ['dict', 'yaml', 'csv', 'table']:
        banner(output)
        print(i.list(format=output))

    banner("reading")
    n = Inventory()
    n.read()
    print(n.list('table'))
    n.save()



def borg(cls):
    cls._state = {}
    orig_init = cls.__init__

    def new_init(self, *args, **kwargs):
        self.__dict__ = cls._state
        orig_init(self, *args, **kwargs)

    cls.__init__ = new_init
    return cls



from __future__ import print_function

from cloudmesh_client.cloud.group import Group
from cloudmesh_client.default import Default
from cloudmesh_client.shell.command import command
from cloudmesh_client.shell.console import Console
from cloudmesh_client.shell.command import PluginCommand, CloudPluginCommand
from cloudmesh_client.common.dotdict import dotdict
from pprint import pprint
from cloudmesh_client.common.hostlist import Parameter


class GroupCommand(PluginCommand, CloudPluginCommand):
    topics = {"group": "cloud"}

    def __init__(self, context):
        self.context = context
        if self.context.debug:
            print("init command group")

    @command
    def do_group(self, args, arguments):
        """
        ::

            Usage:
                group list [GROUPNAME] [--format=FORMAT]
                group remove NAMES [--group=GROUPNAME]
                group add NAMES [--type=TYPE] [--group=GROUPNAME]
                group delete GROUPS
                group copy FROM TO
                group merge GROUPA GROUPB MERGEDGROUP

            manage the groups

            Arguments:

                NAMES        names of object to be added
                GROUPS       names of a groups
                FROM         name of a group
                TO           name of a group
                GROUPA       name of a group
                GROUPB       name of a group
                MERGEDGROUP  name of a group

            Options:
                --format=FORMAT     the output format
                --type=TYPE         the resource type
                --name=NAME         the name of the group
                --id=IDS            the ID(s) to add to the group


            Description:

                Todo: design parameters that are useful and match
                description
                Todo: discuss and propose command

                cloudmesh can manage groups of resource related
                objects. As it would be cumbersome to for example delete
                many virtual machines or delete VMs that are in the same
                group, but are running in different clouds.

                Hence it is possible to add a virtual machine to a
                specific group. The group name to be added to can be set
                as a default. This way all subsequent commands use this
                default group. It can also be set via a command parameter.
                Another convenience function is that the group command can
                use the last used virtual machine. If a vm is started it
                will be automatically added to the default group if it is set.

                If finer grained deletion is needed, it can be achieved
                with the delete command that supports deletion by name

                It is also possible to remove a VM from the group using the
                remove command, by supplying the ID

            Note:

                The type is internally called for the group species, we may
                eliminate the species column and just use the type column for it,

            Example:
                default group mygroup

                group add --type=vm --id=albert-[001-003]
                    adds the vms with the given name using the Parameter
                    see base

                group add --type=vm
                 adds the last vm to the group

                group delete --name=mygroup
                    deletes all objects in the group
        """
        # pprint(arguments)


        if arguments["list"]:

            output = arguments["--format"] or Default.get(name="format", category="general") or "table"
            name = arguments["GROUPNAME"]
            if name is None:

                result = Group.list(output=output)
                if result:
                    print(result)
                else:
                    print("No groups found other than the default group but it has no members.")

            else:

                result = Group.list(name=name,
                                    output=output)

                if result:
                    print(result)
                else:
                    msg_a = ("No group found with name `{name}` found in the "
                             "category `{category}`.".format(**locals()))

                '''
                    # find alternate
                    result = Group.get(name=name)

                    msg_b = ""
                    if result is not None and len(result) < 0:
                        msg_b = " However we found such a variable in " \
                                "category `{category}`. Please consider " \
                                "using --category={category}".format(**locals())
                        Console.error(msg_a + msg_b)
                    else:
                        Console.error("No group with name {name} exists.".format(**locals()))
                '''

                return ""

        elif arguments["add"]:
            # group add NAME... [--type=TYPE] [--category=CLOUD] [--group=GROUP]

            print ("AAA", arguments["NAMES"])
            members = Parameter.expand(arguments["NAMES"])
            print ("MMMM", members)
            data = dotdict({
                "species": arguments["--type"] or "vm",
                "name": arguments["--group"] or Default.group
            })
            print ("DDD", data)
            for member in members:
                data.member = member
                pprint(data)
                Group.add(**data)

            return ""

        elif arguments["delete"]:
            groups = Parameter.expand(arguments["GROUPS"])

            for group in groups:
                result = Group.delete(group)

                if result:
                    Console.ok(result)
                else:
                    Console.error(
                        "delete group {}. failed.".format(group))
            return ""

        elif arguments["remove"]:
            members = Parameter.expand(arguments["NAMES"])

            group = arguments["--group"] or Default.group

            for member in members:
                result = Group.remove(group, member)

                if result:
                    Console.ok(result)
                else:
                    Console.error(
                        "remove {} from group {}. failed.".format(group, member))
            return ""

        elif arguments["copy"]:
            _from = arguments["FROM"]
            _to = arguments["TO"]

            Group.copy(_from, _to)
            return ""

        elif arguments["merge"]:
            _groupA = arguments["GROUPA"]
            _groupB = arguments["GROUPB"]
            _mergedGroup = arguments["MERGEDGROUP"]

            Group.merge(_groupA, _groupB, _mergedGroup)
            return ""



from __future__ import print_function

from cloudmesh_client.cloud.list import List
from cloudmesh_client.shell.command import command
from cloudmesh_client.shell.console import Console

from cloudmesh_client.default import Default
from cloudmesh_client.cloud.image import Image
from cloudmesh_client.cloud.flavor import Flavor
from cloudmesh_client.cloud.vm import Vm
from cloudmesh_client.common.ConfigDict import ConfigDict

from cloudmesh_client.shell.command import PluginCommand, CloudPluginCommand


class ListCommand(PluginCommand, CloudPluginCommand):
    topics = {"list": "cloud"}

    def __init__(self, context):
        self.context = context
        if self.context.debug:
            print("init command list")

    # noinspection PyUnusedLocal
    @command
    def do_list(self, args, arguments):
        """
        ::

            Usage:
                list [--cloud=CLOUD] [--format=FORMAT] [--user=USER] [--tenant=TENANT] default
                list [--cloud=CLOUD] [--format=FORMAT] [--user=USER] [--tenant=TENANT] vm
                list [--cloud=CLOUD] [--format=FORMAT] [--user=USER] [--tenant=TENANT] flavor
                list [--cloud=CLOUD] [--format=FORMAT] [--user=USER] [--tenant=TENANT] image

            List the items stored in the database

            Options:
                --cloud=CLOUD    the name of the cloud
                --format=FORMAT  the output format
                --tenant=TENANT     Name of the tenant, e.g. fg82.

            Description:
                List command prints the values stored in the database
                for [default/vm/flavor/image].
                Result can be filtered based on the cloud, user & tenant arguments.
                If these arguments are not specified, it reads the default

            Examples:
                $ list --cloud india default
                $ list --cloud india --format table flavor
                $ list --cloud india --user albert --tenant fg82 flavor
        """

        # pprint(arguments)

        # Method to get the kind from args
        #
        # TODO: the kind are defined in the provider,
        # TODO: keep the kind lower case
        # why is there a reason to make the gind upper case
        def get_kind():
            for k in ["vm", "image", "flavor", "default"]:
                if arguments[k]:
                    # kinds are all uppercase in model.py
                    return k.upper()
            return "help"

        # Read commandline arguments
        output_format = arguments['--format']
        cloud = arguments['--cloud'] or Default.cloud
        user = arguments['--user']
        tenant = arguments['--tenant']

        # If format is not specified, read default
        if output_format is None:
            output_format = Default.get(name="format") or "table"

        # If cloud is not specified, get default
        if cloud is None:
            cloud = Default.get(name="cloud") or ConfigDict(filename="cloudmesh.yaml")["cloudmesh"]["active"][0]


        # If user is not specified, get default
        if user is None:
            user = Default.get(name="user")

        # If tenant is not specified, get default
        if tenant is None:
            tenant = Default.get(name="tenant")

        # Get the kind
        kind = get_kind()
        header = None
        order = None

        # print help message
        if kind == 'help':
            Console.ok("Print help!")
            return ""

        # Prepare the order & header based on kind
        # TODO: use lower case so we have a convention thats easy to follow
        # TODO: add quota
        # TODO: add limits
        # TODO: add usage

        #
        # TODO: BUG: use the get attribute from the provider.
        # TODO:: call the cm xyz list functions and do not
        # reimplement this here
        # possibly introduce List.py

        result = None
        if kind == 'FLAVOR':
            result = Flavor.list(cloud, format=output_format)
        elif kind == 'DEFAULT':
            result = Default.list(order=order, output=output_format)
        elif kind == 'IMAGE':
            result = Image.list(cloud, format=output_format)
        elif kind == 'VM':
            result = Vm.list(cloud=cloud, output_format=output_format)

        if result:
            print(result)
        else:
            Console.error("No {}s found in the database."
                          .format(kind.lower()))

        return ""



from __future__ import print_function
import requests
import json
import time
from pprint import pprint

# for unit testing only.
USERNAME = ""
PASSWORD = ""


class AuthenticationException(Exception):
    pass


# noinspection PyBroadException
class Authenticator(object):
    base_uri = None
    token = None
    HEADER = {'content-type': 'application/json'}
    # in case of https endpoint
    verify = False

    def __init__(self, uri=None):
        Authenticator.base_uri = uri

    @classmethod
    def set_auth_base_uri(cls, uri):
        cls.base_uri = uri

    @classmethod
    def logon(cls, username, password):
        ret = False
        if not cls.token:
            if cls.base_uri:
                authuri = "%s/login/" % cls.base_uri
                data = {"username": username, "password": password}
                r = requests.post(authuri,
                                  data=json.dumps(data),
                                  headers=cls.HEADER,
                                  verify=cls.verify)
                try:
                    cls.token = r.json()["key"]
                except:
                    ret = False
                ret = cls.token
        else:
            ret = cls.token
        return ret

    @classmethod
    def logoff(cls):
        ret = True
        if cls.token:
            if cls.base_uri:
                authuri = "%s/logout/" % cls.base_uri
                header = cls.HEADER
                header['Authorization'] = "Token %s" % cls.token
                r = requests.post(authuri,
                                  headers=header,
                                  verify=cls.verify)
                cls.token = None
            else:
                ret = False
        return ret

    @classmethod
    def status(cls):
        ret = False
        if cls.token:
            ret = True
        return ret


'''
# To make GET calls for synchronous or asynchronous API
def hop_get(url, headers=None, timeout=10):
    r = requests.get(url, headers=headers)
    ret = None
    # responded immediately
    if r.status_code == 200:
        ret = r.json()
    # code 202, accepted call and processing
    elif r.status_code == 202:
        success = 0
        ntried = 0
        # status check by visiting another URL
        newurl = r.headers["Location"]
        ret = requests.get(newurl, headers=headers).json()
        while not success and ntried<timeout:
            # result ready, status was set to 1; otherwise 0
            success = ret['status']
            if success:
                ret = ret['data']
            else:
                #print "trying again %s out of %s" % (ntried, timeout)
                time.sleep(1)
                ret = requests.get(newurl, headers=headers).json()
                ntried += 1
        if timeout == ntried:
            ret = None
    return ret
'''


# To make GET calls for synchronous or asynchronous API
def hop_get(url, headers=None):
    r = requests.get(url, headers=headers)
    ret = None

    # responded immediately
    if r.status_code == 200:
        ret = r.json()
    # code 202, accepted call and processing
    elif r.status_code == 202:
        # now automatically redirect to result page
        # thus no need to check status periodically.
        # Currently it works well for cluster listing
        # However not sure if the delay is large, what the behaviour would be
        finished = False
        newurl = r.headers["Location"]
        while not finished:
            ret = requests.get(newurl, headers=headers).json()
            # in some occasions, when the result is not ready,
            # the result still has 'status' in it (value as '0')
            # otherwise it's the correct value after redirection
            if 'status' not in ret:
                finished = True
            else:
                time.sleep(1)
    elif r.status_code == 401:
        ret = {"error": "Not Authenticated"}

    return ret


def main():
    url = "http://localhost:8080/rest-auth"
    auth = Authenticator(url)

    # change user, password to proper value as set in django
    # in shell, we may ask user input
    user = USERNAME
    password = PASSWORD
    print(auth.status())
    print(auth.logon(user, password))
    print(auth.status())
    print(auth.logoff())
    print(auth.status())
    print(auth.logoff())


def test_get_cluster_list():
    token = ''
    print("\nTEST 1: Get without logon")
    print("-" * 80)
    authheader = {'content-type': 'application/json',
                  "Authorization": 'Token %s' % token}
    geturl = "http://localhost:8080/v1/cluster/"
    r = requests.get(geturl, headers=authheader)
    pprint(r.json())

    print("\nTEST 2: Auth and then get cluster list")
    print("-" * 80)
    authurl = "http://localhost:8080/rest-auth"
    auth = Authenticator(authurl)
    # change user, password to proper value as set in django
    # in shell, we may ask user input
    user = USERNAME
    password = PASSWORD
    token = auth.logon(user, password)

    # construct a header with auth token after login
    # for all the following calls before log out
    authheader = {'content-type': 'application/json',
                  "Authorization": 'Token %s' % token}
    geturl = "http://localhost:8080/v1/cluster/"
    r = hop_get(geturl, headers=authheader)
    pprint(r)

    # as of 2:40pm ET Oct 15, this is changed to 'not implemented'
    # as of 5:30pm ET this is now fixed and working
    print("\nTEST 3: Get cluster 'OSG'")
    print("-" * 80)
    geturl1 = "%s%s" % (geturl, "osg/")
    r1 = hop_get(geturl1, headers=authheader)
    pprint(r1)

    print("\nTEST 4: logoff and get cluster list again")
    print("-" * 80)
    auth.logoff()
    authheader = {'content-type': 'application/json',
                  "Authorization": 'Token %s' % token}
    geturl = "http://localhost:8080/v1/cluster/"
    r = requests.get(geturl, headers=authheader)
    pprint(r.json())


if __name__ == "__main__":
    test_get_cluster_list()
    # main()



from __future__ import print_function

from cloudmesh_client.common.util import path_expand
from cloudmesh_client.shell.command import command
from cloudmesh_client.shell.console import Console

# from cloudmesh_client.shell.cm import c
import os
from cloudmesh_client.shell.command import PluginCommand, CloudPluginCommand


class ResetCommand(PluginCommand, CloudPluginCommand):
    topics = {"reset": "cloud"}

    def __init__(self, context):
        self.context = context
        if self.context.debug:
            print("init command reset")

    # noinspection PyUnusedLocal
    @command
    def do_reset(self, args, arguments):
        """
        ::

          Usage:
              reset

        Description:

            DANGER: This method erases the database.


        Examples:
            clean

        """
        filename = path_expand("~/.cloudmesh/cloudmesh.db")
        if os.path.exists(filename):
            os.remove(filename)
        Console.ok("Database reset")
        #r = self.do_register("profile")
        #r = self.do_info("")

        r = self.do_quit(None)
        Console.error(
            "Quitting the shell does not yet work. please exit the shell now.")
        return ""




from cloudmesh_client.shell.console import Console
from cloudmesh_client.shell.command import command
from cloudmesh_client.shell.command import PluginCommand

"""



            ARGUMENTS:
                FILENAME  the file to open in the cwd if . is
                          specified. If file in in cwd
                          you must specify it with ./FILENAME
"""


# noinspection PyUnusedLocal,PyBroadException
class WhoCommand(PluginCommand):
    topics = {"Who ": "system"}

    def __init__(self, context):
        self.context = context
        self.context.who_token = None
        if self.context.debug:
            Console.ok("init Who command")

    @command
    def do_who(self, args, arguments):
        """
        ::

            Usage:
               who  hostname

        """

        try:
            logon = who.logon()
            if logon is False:
                Console.error("Could not logon")
                return
        except:
            Console.error("Could not logon")
        # pprint (arguments)
        output_format = arguments["--format"] or "table"

        if arguments["status"]:
            pass

        ValueError("NOT yet implemented")

        return ""



class Attributes(object):
    @classmethod
    def get(cls, kind):
        layout = {
            'default': {
                'order': ['user',
                          'category',
                          'name',
                          'value',
                          'updated_at'],
                'header': ['User',
                           'Category',
                           'Name',
                           'Value',
                           'Updated']
            },
            'var': {
                'order': ['user',
                          'name',
                          'value',
                          'updated_at'],
                'header': ['User',
                           'Name',
                           'Value',
                           'Updated']
            },
        }
        if kind in layout:
            order = layout[kind]['order']
            header = layout[kind]['header']
        else:
            order = None
            header = None

        return order, header



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Changes:
# * June 2015, Gregor von Laszewski, using new style raise of exceptions
# * June 2015, We assume we can redictibute it under LGPL license, we contacted the authors
#              but have not yet heard back from them. We asked if we can distribute under
#              apache license.
# * June 2015  added class wrapper Parameter
#
# Hostlist library
#
# Copyright (C) 2008 Kent Engström <kent@nsc.liu.se>,
#                    Thomas Bellman <bellman@nsc.liu.se> and
#                    Pär Andersson <paran@nsc.liu.se>,
#                    National Supercomputer Centre
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

"""Handle hostlist expressions.
This module provides operations to expand and collect hostlist
expressions.
The hostlist expression syntax is the same as in several programs
developed at LLNL (https://computing.llnl.gov/linux/). However in
corner cases the behaviour of this module have not been compared for
compatibility with pdsh/dshbak/SLURM et al.
"""

__version__ = "1.14"

import re
import itertools


# Exception used for error reporting to the caller
class BadHostlist(Exception): pass


# Configuration to guard against ridiculously long expanded lists
MAX_SIZE = 100000


# Hostlist expansion

class Parameter(object):
    @classmethod
    def expand(cls, parameter, allow_duplicates=False, sort=False):
        return expand_hostlist(parameter, allow_duplicates=False, sort=False)

    '''
    def expand(cls, *args, **kwargs):

        allow_duplicates = False
        sort = False

        if "allow_duplicates" in kwargs:
            allow_duplicates = kwargs["allow_duplicates"]

        if "sort" in kwargs:
            sort = kwargs["sort"]

        parameters = ','.join(*args)

        return expand_hostlist(parameters, allow_duplicates=allow_duplicates, sort=sort)
    '''


def expand_hostlist(hostlist, allow_duplicates=False, sort=False):
    """Expand a hostlist expression string to a Python list.
    Example: expand_hostlist("n[9-11],d[01-02]") ==>
             ['n9', 'n10', 'n11', 'd01', 'd02']
    Unless allow_duplicates is true, duplicates will be purged
    from the results. If sort is true, the output will be sorted.
    """

    results = []
    bracket_level = 0
    part = ""

    for c in hostlist + ",":
        if c == "," and bracket_level == 0:
            # Comma at top level, split!
            if part: results.extend(expand_part(part))
            part = ""
            bad_part = False
        else:
            part += c

        if c == "[":
            bracket_level += 1
        elif c == "]":
            bracket_level -= 1

        if bracket_level > 1:
            raise BadHostlist("nested brackets")
        elif bracket_level < 0:
            raise BadHostlist("unbalanced brackets")

    if bracket_level > 0:
        raise BadHostlist("unbalanced brackets")

    if not allow_duplicates:
        results = remove_duplicates(results)
    if sort:
        results = numerically_sorted(results)
    return results


def expand_part(s):
    """Expand a part (e.g. "x[1-2]y[1-3][1-3]") (no outer level commas)."""

    # Base case: the empty part expand to the singleton list of ""
    if s == "":
        return [""]

    # Split into:
    # 1) prefix string (may be empty)
    # 2) rangelist in brackets (may be missing)
    # 3) the rest

    m = re.match(r'([^,\[]*)(\[[^\]]*\])?(.*)', s)
    (prefix, rangelist, rest) = m.group(1, 2, 3)

    # Expand the rest first (here is where we recurse!)
    rest_expanded = expand_part(rest)

    # Expand our own part
    if not rangelist:
        # If there is no rangelist, our own contribution is the prefix only
        us_expanded = [prefix]
    else:
        # Otherwise expand the rangelist (adding the prefix before)
        us_expanded = expand_rangelist(prefix, rangelist[1:-1])

    # Combine our list with the list from the expansion of the rest
    # (but guard against too large results first)
    if len(us_expanded) * len(rest_expanded) > MAX_SIZE:
        raise BadHostlist("results too large")

    return [us_part + rest_part
            for us_part in us_expanded
            for rest_part in rest_expanded]


def expand_rangelist(prefix, rangelist):
    """ Expand a rangelist (e.g. "1-10,14"), putting a prefix before."""

    # Split at commas and expand each range separately
    results = []
    for range_ in rangelist.split(","):
        results.extend(expand_range(prefix, range_))
    return results


def expand_range(prefix, range_):
    """ Expand a range (e.g. 1-10 or 14), putting a prefix before."""

    # Check for a single number first
    m = re.match(r'^[0-9]+$', range_)
    if m:
        return ["%s%s" % (prefix, range_)]

    # Otherwise split low-high
    m = re.match(r'^([0-9]+)-([0-9]+)$', range_)
    if not m:
        raise BadHostlist("bad range")

    (s_low, s_high) = m.group(1, 2)
    low = int(s_low)
    high = int(s_high)
    width = len(s_low)

    if high < low:
        raise BadHostlist("start > stop")
    elif high - low > MAX_SIZE:
        raise BadHostlist("range too large")

    results = []
    for i in xrange(low, high + 1):
        results.append("%s%0*d" % (prefix, width, i))
    return results


def remove_duplicates(l):
    """Remove duplicates from a list (but keep the order)."""
    seen = set()
    results = []
    for e in l:
        if e not in seen:
            results.append(e)
            seen.add(e)
    return results


# Hostlist collection

def collect_hostlist(hosts, silently_discard_bad=False):
    """Collect a hostlist string from a Python list of hosts.
    We start grouping from the rightmost numerical part.
    Duplicates are removed.
    A bad hostname raises an exception (unless silently_discard_bad
    is true causing the bad hostname to be silently discarded instead).
    """

    # Split hostlist into a list of (host, "") for the iterative part.
    # (Also check for bad node names now)
    # The idea is to move already collected numerical parts from the
    # left side (seen by each loop) to the right side (just copied).

    left_right = []
    for host in hosts:
        # We remove leading and trailing whitespace first, and skip empty lines
        host = host.strip()
        if host == "": continue

        # We cannot accept a host containing any of the three special
        # characters in the hostlist syntax (comma and flat brackets)
        if re.search(r'[][,]', host):
            if silently_discard_bad:
                continue
            else:
                raise BadHostlist("forbidden character")

        left_right.append((host, ""))

    # Call the iterative function until it says it's done
    looping = True
    while looping:
        left_right, looping = collect_hostlist_1(left_right)
    return ",".join([left + right for left, right in left_right])


def collect_hostlist_1(left_right):
    """Collect a hostlist string from a list of hosts (left+right).
    The input is a list of tuples (left, right). The left part
    is analyzed, while the right part is just passed along
    (it can contain already collected range expressions).
    """

    # Scan the list of hosts (left+right) and build two things:
    # *) a set of all hosts seen (used later)
    # *) a list where each host entry is preprocessed for correct sorting

    sortlist = []
    remaining = set()
    for left, right in left_right:
        host = left + right
        remaining.add(host)

        # Match the left part into parts
        m = re.match(r'^(.*?)([0-9]+)?([^0-9]*)$', left)
        (prefix, num_str, suffix) = m.group(1, 2, 3)

        # Add the right part unprocessed to the suffix.
        # This ensures than an already computed range expression
        # in the right part is not analyzed again.
        suffix = suffix + right

        if num_str is None:
            # A left part with no numeric part at all gets special treatment!
            # The regexp matches with the whole string as the suffix,
            # with nothing in the prefix or numeric parts.
            # We do not want that, so we move it to the prefix and put
            # None as a special marker where the suffix should be.
            assert prefix == ""
            sortlist.append(((host, None), None, None, host))
        else:
            # A left part with at least an numeric part
            # (we care about the rightmost numeric part)
            num_int = int(num_str)
            num_width = len(num_str)  # This width includes leading zeroes
            sortlist.append(((prefix, suffix), num_int, num_width, host))

    # Sort lexicographically, first on prefix, then on suffix, then on
    # num_int (numerically), then...
    # This determines the order of the final result.

    sortlist.sort()

    # We are ready to collect the result parts as a list of new (left,
    # right) tuples.

    results = []
    needs_another_loop = False

    # Now group entries with the same prefix+suffix combination (the
    # key is the first element in the sortlist) to loop over them and
    # then to loop over the list of hosts sharing the same
    # prefix+suffix combination.

    for ((prefix, suffix), group) in itertools.groupby(sortlist,
                                                       key=lambda x: x[0]):

        if suffix is None:
            # Special case: a host with no numeric part
            results.append(("", prefix))  # Move everything to the right part
            remaining.remove(prefix)
        else:
            # The general case. We prepare to collect a list of
            # ranges expressed as (low, high, width) for later
            # formatting.
            range_list = []

            for ((prefix2, suffix2), num_int, num_width, host) in group:
                if host not in remaining:
                    # Below, we will loop internally to enumate a whole range
                    # at a time. We then remove the covered hosts from the set.
                    # Therefore, skip the host here if it is gone from the set.
                    continue
                assert num_int is not None

                # Scan for a range starting at the current host
                low = num_int
                while True:
                    host = "%s%0*d%s" % (prefix, num_width, num_int, suffix)
                    if host in remaining:
                        remaining.remove(host)
                        num_int += 1
                    else:
                        break
                high = num_int - 1
                assert high >= low
                range_list.append((low, high, num_width))

            # We have a list of ranges to format. We make sure
            # we move our handled numerical part to the right to
            # stop it from being processed again.
            needs_another_loop = True
            if len(range_list) == 1 and range_list[0][0] == range_list[0][1]:
                # Special case to make sure that n1 is not shown as n[1] etc
                results.append((prefix,
                                "%0*d%s" %
                                (range_list[0][2], range_list[0][0], suffix)))
            else:
                # General case where high > low
                results.append((prefix, "[" + \
                                ",".join([format_range(l, h, w)
                                          for l, h, w in range_list]) + \
                                "]" + suffix))

    # At this point, the set of remaining hosts should be empty and we
    # are ready to return the result, together with the flag that says
    # if we need to loop again (we do if we have added something to a
    # left part).
    assert not remaining
    return results, needs_another_loop


def format_range(low, high, width):
    """Format a range from low to high inclusively, with a certain width."""

    if low == high:
        return "%0*d" % (width, low)
    else:
        return "%0*d-%0*d" % (width, low, width, high)


# Sort a list of hosts numerically

def numerically_sorted(l):
    """Sort a list of hosts numerically.
    E.g. sorted order should be n1, n2, n10; not n1, n10, n2.
    """

    return sorted(l, key=numeric_sort_key)


nsk_re = re.compile("([0-9]+)|([^0-9]+)")


def numeric_sort_key(x):
    return [handle_int_nonint(i_ni) for i_ni in nsk_re.findall(x)]


def handle_int_nonint(int_nonint_tuple):
    if int_nonint_tuple[0]:
        return int(int_nonint_tuple[0])
    else:
        return int_nonint_tuple[1]


# Parse SLURM_TASKS_PER_NODE into a list of task numbers
#
# Description from the SLURM sbatch man page:
#              Number of tasks to be initiated on each node. Values
#              are comma separated and in the same order as
#              SLURM_NODELIST.  If two or more consecutive nodes are
#              to have the same task count, that count is followed by
#              "(x#)" where "#" is the repetition count. For example,
#              "SLURM_TASKS_PER_NODE=2(x3),1" indicates that the first
#              three nodes will each execute three tasks and the
#              fourth node will execute one task.

def parse_slurm_tasks_per_node(s):
    res = []
    for part in s.split(","):
        m = re.match(r'^([0-9]+)(\(x([0-9]+)\))?$', part)
        if m:
            tasks = int(m.group(1))
            repetitions = m.group(3)
            if repetitions is None:
                repetitions = 1
            else:
                repetitions = int(repetitions)
            if repetitions > MAX_SIZE:
                raise BadHostlist("task list repetitions too large")
            for i in range(repetitions):
                res.append(tasks)
        else:
            raise BadHostlist("bad task list syntax")
    return res


#
# Keep this part to tell users where the command line interface went
#

if __name__ == '__main__':
    import os, sys

    sys.stderr.write("The command line utility has been moved to a separate 'hostlist' program.\n")
    sys.exit(os.EX_USAGE)



from __future__ import print_function
from cloudmesh_client.shell.command import command
from cloudmesh_client.shell.console import Console
from cloudmesh_client.default import Default
from cloudmesh_client.shell.command import PluginCommand, CloudPluginCommand
from cloudmesh_client.default import Default
from cloudmesh_client.common.Printer import Printer
from cloudmesh_client.common.dotdict import dotdict


class InfoCommand(PluginCommand, CloudPluginCommand):
    topics = {"info": "cloud"}

    def __init__(self, context):
        self.context = context
        if self.context.debug:
            print("init command info")

    # noinspection PyUnusedLocal
    @command
    def do_info(self, args, arguments):
        """
        ::

            Usage:
                info [--cloud=CLOUD] [--format=FORMAT]

            Options:
               --format=FORMAT  the output format [default: table]
               --cloud=CLOUD    the cloud name

            Examples:
                cm info

        """
        arg = dotdict(arguments)
        arg.cloud = arguments["--cloud"] or Default.cloud
        arg.FORMAT = arguments["--format"] or "table"

        d = {
            "cloud": arg.cloud,
            "key": Default.key,
            "user": Default.user,
            "vm": Default.vm,
            "group": Default.group,
            "secgroup": Default.secgroup,
            "counter": Default.get_counter(name="name"),
            "image": Default.get_image(category=arg.cloud),
            "flavor": Default.get_flavor(category=arg.cloud),
            "refresh": str(Default.refresh),
            "debug": str(Default.debug),
            "interactive": str(Default.interactive),
            "purge": str(Default.purge),

        }
        order = ["cloud", "key", "user", "vm", "group", "secgroup",
                 "counter", "image", "flavor", "refresh", "debug", "interactive", "purge"]
        print(Printer.attribute(d, order=order, output=arg.FORMAT, sort_keys=False))

        if d["key"] in ["TBD", ""] or d["user"] in ["TBD", ""]:
            msg = "Please replace the TBD values"
            msg = msg + "\nSee Also: \n\n" \
                  + "    cm register profile \n" \
                  + "    cm default user=YOURUSERNAME\n"
            Console.error(msg, traceflag=False)
        return ""



from cloudmesh_client.cloud.network import Network

from cloudmesh_client.cloud.image import Image

from cloudmesh_client.cloud.vm import Vm

from cloudmesh_client import Console
from cloudmesh_client.api import Resource, Provider, Node

from cloudmesh_client.db.openstack.model import IMAGE_OPENSTACK, VM_OPENSTACK
from cloudmesh_client.common.util import exponential_backoff
from cloudmesh_client.default import Default


class ImageResource(Resource):
    def create(self):
        pass

    def delete(self):
        pass

    def list(self):
        pass

    def refresh(self):
        pass



class KeyResource(Resource):
    pass


class FloatingIpResource(Resource):
    pass


class FlavorResource(Resource):
    pass


class OpenstackProvider(Provider):

    # def __init__(self):
    #     super(OpenstackProvider, self).__init__()
    #
    #     self.images = ImageResource()
    #     self._add_resource(self.images)
    #
    #     self.keys = KeyResource()
    #     self._add_resource(self.keys)
    #
    #     self.flavors = FlavorResource()
    #     self._add_resource(self.flavors)
    #
    #     self.floating_ips = FloatingIpResource()
    #     self._add_resource(self.floating_ips)


    def boot(self, **kwargs):
        """Boot a single VM

        :param kwargs: parameters to :meth:`Vm.boot`
        :return: the vm details
        :rtype: :class:`Node`
        """

        cloud = kwargs.get('cloud', Default.cloud)
        name = kwargs.get('name', Vm.generate_vm_name())
        image = kwargs.get('image', Default.image)
        flavor = kwargs.get('flavor', Default.flavor)
        key = kwargs.get('key', Default.key)
        secgroup = kwargs.get('secgroup', Default.secgroup)
        group = kwargs.get('group', Default.group)
        username = kwargs.get('username', Image.guess_username(image))
        cluster = kwargs.get('cluster', None)

        # shorthand for getting a dict of all the vm details
        #
        # IMPORTANT: anything declared prior to the call to `locals()`
        # may be passed to `Vm.boot`, so make sure that only parameters are
        # defined above this comment.
        details = locals()
        details.pop('kwargs')

        # currently, Vm.boot returns the instance UUID from the provider for openstack images
        # 2016/12/12
        uuid = Vm.boot(**details)


        # helper function: the Vm.boot only returns a UUID, but we
        # need to use the VM model instead. Additionally, we'll need
        # to poll the VM to wait until it is active.
        #
        # The kwargs are used to select the item from the DB:
        # eg: uuid=???, cm_id=???, etc
        def get_vm(**kwargs):
            """Selects the VM based on the given properties"""
            model = self.db.vm_table_from_provider('openstack')
            vm = self.db.select(model, **kwargs).all()
            assert len(vm) == 1, vm
            vm = vm[0]
            return vm

        # get the VM from the UUID
        vm = get_vm(uuid=uuid)
        cm_id = vm.cm_id

        def is_active():
            Vm.refresh(cloud=cloud)
            vm = get_vm(cm_id=cm_id)
            return vm.status == 'ACTIVE'

        if not exponential_backoff(is_active):
            Console.error('Failed to get ACTIVE vm within timeframe')
            raise ValueError

        assert is_active()
        vm = get_vm(cm_id=cm_id)
        assert isinstance(vm, VM_OPENSTACK), vm.__class__

        return OpenstackNode(model=vm, provider=self)

    def create_ip(self, node):

        ip = Network.find_assign_floating_ip(
            cloudname=self.cloud,
            instance_id=node.name,
        )

        Vm.refresh(cloud=self.cloud)

        Console.ok('Assigned ip to {}: {}'.format(node.name, ip))

    def delete(self, nodde):
        raise NotImplementedError

    def node(self):
        raise NotImplementedError


class OpenstackNode(Node):

    def __init__(self, model, provider):
        super(Node, self).__init__()
        self._model = model
        self._provider = provider

    @property
    def name(self):
        return self._model.name

    @property
    def username(self):
        return self._model.username

    @property
    def private_ip(self):
        return self.model.static_ip

    @property
    def public_ip(self):
        return self.model.floating_ip

    def boot(self, **kwargs):
        pass

    def delete(self):
        raise NotImplementedError

    def start(self):
        raise NotImplementedError

    def stop(self):
        raise NotImplementedError

    def ssh(self, cmd=None, user=None):
        raise NotImplementedError

    def create_ip(self):
        self._provider.create_ip(self)



from __future__ import print_function

import time

from cloudmesh_client.cloud.iaas.CloudProvider import CloudProvider
from cloudmesh_client.cloud.network import Network
from cloudmesh_client.cloud.vm import Vm
from cloudmesh_client.db import CloudmeshDatabase, IntegrityError
from cloudmesh_client.db.general.model import CLUSTER
from cloudmesh_client.default import Default
from cloudmesh_client.exc import ClusterNameClashException
from cloudmesh_client.shell.console import Console


_db = CloudmeshDatabase()


# FIXME: refactor with Vm.generate_vm_name
# duplicate functionality
# check cloudmesh_client.cloud.counter.Counter
def generate_cluster_name(prefix=None, offset=0, fill=3):
    prefix = (prefix + '-') if prefix else ''

    counter_name = 'cluster'
    counter = Default.get_counter(counter_name)
    Default.set_counter(counter_name, counter + 1)

    index = str(counter).zfill(fill)
    name = prefix + 'cluster' + '-' + index

    return name


class Cluster(CLUSTER):

    def __init__(self, *args, **kwargs):
        super(Cluster, self).__init__(*args, **kwargs)

        self.provider = CloudProvider(self.cloud).provider.cloud_type

        try:
            _db.insert(self)
        except IntegrityError as e:
            line = 'UNIQUE constraint failed: {}.name'\
                   .format(self.__tablename__)
            if line in e.message:
                raise ClusterNameClashException(self.__tablename__, self.name)
            else:
                raise

    @property
    def instances(self):

        tablename = 'vm_{}'.format(self.provider)
        table = _db.table(name=tablename)
        return _db.select(table, cluster=self.name).all()

    def __len__(self):
        return len(self.instances)

    def _assign_floating_ip(self, vm):
        """Assign a floating ip

        :returns: floating ip
        :rtype: :class:`str`
        """

        ip = Network.find_assign_floating_ip(
            cloudname=self.cloud,
            instance_id=vm.name,
        )

        Vm.refresh(cloud=self.cloud)

        Console.ok('Assigned ip to {}: {}'.format(vm.name, ip))

        return ip

    def assign_floating_ip(self):
        """Assign a floating ip to all nodes in the cluster, if they do not
        already have one
        """

        for vm in self.instances:
            if not vm.floating_ip:
                self._assign_floating_ip(vm)

    def _wait_until_active(self, vm, sleeptime_s=5):
        Console.info('Waiting until {} is active'.format(vm.name))

        while True:
            Vm.refresh(cloud=self.cloud)
            vm = self.get_vm(cm_id=vm.cm_id)

            if vm.status == 'ACTIVE':
                return
            else:
                Console.debug_msg('Status is: {}'.format(vm.status))
                Console.debug_msg('Waiting for {} seconds'.format(sleeptime_s))
                time.sleep(sleeptime_s)

    def get_vm(self, **kwargs):
        model = _db.table(name='vm_{}'.format(self.provider))
        vm = _db.select(model, **kwargs).all()
        assert len(vm) == 1, vm
        vm = vm[0]
        return vm

    def boot_single(self):
        """Boots a new instance and adds it to this cluster
        """

        instance_name = Vm.generate_vm_name()

        uuid = Vm.boot(
            cloud=self.cloud,
            key=self.key,
            name=instance_name,
            image=self.image,
            flavor=self.flavor,
            group=self.secgroup,
            cluster=self.name,
        )

        vm = self.get_vm(uuid=uuid)
        self._wait_until_active(vm)

        if self.assignFloatingIP:
            vm.floating_ip = self._assign_floating_ip(vm)

    def boot(self):
        """Boot all the nodes in the cluster
        """

        for _ in xrange(self.count - len(self.instances)):
            self.boot_single()

    def delete(self, force=False):
        """Delete this cluster and all component nodes
        """

        for node in self.instances:
            Vm.delete(servers=[node.name], force=force)

        _db.delete_(self.__class__, cm_id=self.cm_id)



""" run with

python setup.py install; nosetests -v --nocapture  tests/cm_hpc/test_hpc.py:Test_hpc.test_001

python setup.py install; nosetests -v --nocapture tests/cm_hpc/test_hpc.py

or

nosetests -v tests/test_hpc.py

"""

from cloudmesh_client.common.Shell import Shell
from cloudmesh_client.common.dotdict import dotdict
from cloudmesh_client.common.util import HEADING
from cloudmesh_client.common.util import banner


# noinspection PyPep8Naming
class Test_hpc:
    """
        This class tests the HpcCommand
    """
    # noinspection PyTypeChecker
    data = dotdict({
        "cluster": "comet",
    })

    def run(self, command):
        command = command.format(**self.data)
        banner(command, c="-")
        print(command)
        parameter = command.split(" ")
        shell_command = parameter[0]
        args = parameter[1:]
        result = Shell.execute(shell_command, args)
        print(result)
        return str(result)

    # noinspection PyPep8Naming
    def tearDown(self):
        pass

    def setup(self):
        HEADING("set default cluster")
        result = self.run("cm default cluster={cluster}")
        assert "{cluster}".format(**self.data) in result

    def test_001(self):
        HEADING("test hpc info")
        result = self.run("cm hpc info --cluster={cluster}")
        assert "{cluster}".format(**self.data) in result

    def test_002(self):
        HEADING("test hpc queue ")
        result = self.run("cm hpc queue --cluster={cluster}")
        assert "{cluster}".format(**self.data) in result

    def test_003(self):
        HEADING(" test hpc status  ")
        result = self.run("cm hpc status --cluster={cluster}")
        assert "{cluster}".format(**self.data) in result



import inspect
import textwrap
import shlex

from docopt import docopt


class PluginCommand(object):
    pass


class CloudPluginCommand(object):
    pass


class ShellPluginCommand(object):
    pass


class HPCPluginCommand(object):
    pass


class CometPluginCommand(object):
    pass


# noinspection PySingleQuotedDocstring
def command(func):
    '''
    A decorator to create a function with docopt arguments.
    It also generates a help function

    @command
    def do_myfunc(self, args):
        """ docopts text """
        pass

    will create

    def do_myfunc(self, args, arguments):
        """ docopts text """
        ...

    def help_myfunc(self, args, arguments):
        ... prints the docopt text ...

    :param func: the function for the decorator
    '''
    classname = inspect.getouterframes(inspect.currentframe())[1][3]
    name = func.__name__
    help_name = name.replace("do_", "help_")
    doc = textwrap.dedent(func.__doc__)

    def new(instance, args):
        # instance.new.__doc__ = doc
        try:
            argv = shlex.split(args)
            arguments = docopt(doc, help=True, argv=argv)
            func(instance, args, arguments)
        except SystemExit as e:
            if args not in ('-h', '--help'):
                print("Could not execute the command.")
                print(e)
            print(doc)

    new.__doc__ = doc
    return new



import webbrowser
import os

from cloudmesh_client.shell.console import Console
from cloudmesh_client.shell.command import command
from cloudmesh_client.shell.command import PluginCommand, ShellPluginCommand, \
    CometPluginCommand


# noinspection PyUnusedLocal,PyBroadException,PyBroadException
class OpenCommand(PluginCommand, ShellPluginCommand, CometPluginCommand):
    topics = {"open": "shell"}

    def __init__(self, context):
        self.context = context
        if self.context.debug:
            print("init command browser")

    def _expand_filename(self, line):
        """expands the filename if there is a . as leading path"""
        # expand .
        newline = line
        path = os.getcwd()
        if newline.startswith("."):
            newline = newline.replace(".", path, 1)
        # expand ~
        newline = os.path.expanduser(newline)
        return newline

    @command
    def do_open(self, args, arguments):
        """
        ::

            Usage:
                    open FILENAME

            ARGUMENTS:
                FILENAME  the file to open in the cwd if . is
                          specified. If file in in cwd
                          you must specify it with ./FILENAME

            Opens the given URL in a browser window.
        """
        filename = arguments['FILENAME']
        filename = self._expand_filename(filename)
        Console.ok("open {0}".format(filename))

        if not (filename.startswith("file:") or filename.startswith("http:")):
            try:
                with open(filename):
                    pass
                filename += "file://"
            except:
                Console.error(
                    "unsupported browser format in file {0}".format(filename))
                return ""

        try:
            webbrowser.open("%s" % filename)
        except:
            Console.error(
                "can not open browser with file {0}".format(filename))



from __future__ import print_function
from cloudmesh_client.shell.console import Console
from cloudmesh_client.shell.command import command, PluginCommand, \
    ShellPluginCommand, CometPluginCommand
from cloudmesh_client.common.ConfigDict import ConfigDict


class ColorCommand(PluginCommand, ShellPluginCommand, CometPluginCommand):
    topics = {"color": "shell"}

    def __init__(self, context):
        self.context = context
        if self.context.debug:
            print("init command color")

    # noinspection PyUnusedLocal
    @command
    def do_color(self, args, arguments):
        """
        ::

            Usage:
                color FLAG

            Arguments:

                FLAG    color mode flag ON/OFF

            Description:

                Global switch for the console color mode.
                One can switch the color mode on/off with
                    cm color ON
                    cm color OFF

                By default, the color mode is ON

            Examples:
                color ON
                color OFF
        """

        # default mode is ON
        color_mode = True

        flag = arguments["FLAG"].lower()
        if flag in ["on", "true"]:
            color_mode = True
            Console.set_theme(color=True)
        elif flag in ["off", "false"]:
            color_mode = False
            Console.set_theme(color=False)
        else:
            Console.error("Invalid Flag")
            return

        # Update the cloudmesh.yaml file
        config = ConfigDict("cloudmesh.yaml")
        config["cloudmesh"]["system"]["console_color"] = color_mode
        config.save()

        Console.color = color_mode
        Console.ok("Color {:}".format(str(color_mode)))

        return ""



from cloudmesh_client.cloud.iaas.provider.openstack.CloudProviderOpenstackAPI import \
    set_os_environ
from cloudmesh_client.common.Printer import Printer
from cloudmesh_client.common.Shell import Shell
from cloudmesh_client.cloud.ListResource import ListResource
from cloudmesh_client.common.TableParser import TableParser
from cloudmesh_client.cloud.nova import Nova


# Note: This command is currently implemented using the nova command as we
# have not yet found how to print it from the api. We may want to do a rest
# call instead?

class Usage(ListResource):
    @classmethod
    def list(cls, cloud, start=None, end=None, tenant=None, format="table"):
        # set the environment variables
        set_os_environ(cloud)

        try:
            # execute the command
            args = ["usage"]
            if start is not None:
                args.extend(["--start", start])
            if end is not None:
                args.extend(["--end", end])
            if tenant is not None:
                args.extend(["--tenant", tenant])

            result = Shell.execute("nova", args)
            result = Nova.remove_subjectAltName_warning(result)

            lines = result.splitlines()
            dates = lines[0]

            # TODO: as stated below, nova returns additional lines,
            # on my pc, SecurityWarning is returned, so filtering..

            for l in lines[1:]:
                if l.__contains__("SecurityWarning"):
                    lines.remove(l)

            table = '\n'.join(lines[1:])

            dates = dates.replace("Usage from ", "").replace("to", "").replace(" +", " ")[:-1].split()

            #
            # TODO: for some reason the nova command has returned not the
            # first + char, so we could not ignore the line we may set - as
            # additional comment char, but that did not work
            #

            d = TableParser.convert(table, comment_chars="+#")

            # d["0"]["start"] = "start"
            # d["0"]["end"] = "end"

            d["0"]["start"] = dates[0]
            d["0"]["end"] = dates[1]

            # del d['0']

            return Printer.write(d,
                                 order=["start",
                                        "end",
                                        "servers",
                                        "cpu hours",
                                        "ram mb-hours",
                                        "disk gb-hours"],
                                 output=format)

        except Exception as e:
            return e



from pprint import pprint
from cloudmesh_client.common.FlatDict2 import FlatDict2


class AzureDict(object):

    @classmethod
    def convert_to_vm_dict(cls, hosted_service_obj, deployment_obj):
        # pprint("In convert_to_vm_dict")
        vm_dict = dict()

        # option1: fetching from the objects directly
        vm_dict['id'] = deployment_obj.private_id
        vm_dict['name'] = deployment_obj.name
        vm_dict['instance_name'] = deployment_obj.name
        vm_dict['cloud_service'] = hosted_service_obj.service_name
        vm_dict['status'] = deployment_obj.status
        vm_dict['dns_name'] = deployment_obj.url
        if deployment_obj.virtual_ips is not None and len(deployment_obj.virtual_ips.virtual_ips) > 0:
            vm_dict['public_ips'] = deployment_obj.virtual_ips.virtual_ips[0].address
            # pprint((deployment_obj.virtual_ips.virtual_ips)[0].address)

        # option2: Using dict to fetch
        hosted_service_dict = FlatDict2.convert(hosted_service_obj)
        deployment_dict = FlatDict2.convert(deployment_obj)
        # pprint(deployment_dict)

        # vm_dict['id'] = deployment_dict['private_id']
        # vm_dict['name'] = deployment_dict['name']
        # vm_dict['instance_name'] = deployment_dict['name']
        # vm_dict['cloud_service'] = hosted_service_dict['service_name']
        # vm_dict['status'] = deployment_dict['status']
        # vm_dict['dns_name'] = deployment_dict['url']
        # if 'virtual_ips__virtual_ips' in deployment_dict and len(deployment_dict['virtual_ips__virtual_ips']) > 0:
        #     vm_dict['public_ips'] = deployment_dict['virtual_ips__virtual_ips'][0]['address']
        # if 'role_instance_list__role_instances' in deployment_dict and len(deployment_dict['role_instance_list__role_instances']) > 0:
        #     vm_dict['private_ips'] = deployment_dict['role_instance_list__role_instances'][0]['ip_address']
        # if 'role_list__roles' in deployment_dict and len(deployment_dict['role_list__roles']) > 0:
        #     if 'os_virtual_hard_disk' in deployment_dict['role_list__roles'][0]:
        #         vm_dict['image_name'] = deployment_dict['role_list__roles'][0]['os_virtual_hard_disk']['source_image_name']

        # vm_dict['resource_location'] = hosted_service_dict['hosted_service_properties']['hosted_service_properties']['ResourceLocation']
        return vm_dict


    @classmethod
    def convert_to_image_dict(cls, image_obj):
        image_dict = dict()
        image_dict['id'] = image_obj.name
        image_dict_values = FlatDict2.convert(image_obj)
        for key in image_dict_values:
            image_dict[key] = image_dict_values[key]
        return image_dict


    @classmethod
    def convert_to_flavor_dict(cls, flavor_obj):
        flavor_dict = dict()
        flavor_dict['id'] = flavor_obj.name
        flavor_dict_values = FlatDict2.convert(flavor_obj)
        for key in flavor_dict_values:
            flavor_dict[key] = flavor_dict_values[key]
        return flavor_dict




import inspect
import logging

from cloudmesh_client.common.ConfigDict import ConfigDict
from cloudmesh_client.common.util import path_expand
from cloudmesh_client.default import Default

# define the logger
LOGGER = logging.getLogger('LogUtil')


class LogUtil(object):
    # define global format for logs
    FORMAT = "%(asctime)s [%(levelname)s] %(filename)s:%(lineno)s %(funcName)s() %(message)s"

    # set global key for log
    LOG_LEVEL_KEY = "log_level"

    # default log level is ERROR
    DEFAULT_LOG_LEVEL = "ERROR"

    category = "general"

    @staticmethod
    def save():
        """
        save the loglevel for a cloud to the cloudmesh.yaml file
        """
        # TODO: BUG: this seems inconsistant as loglevels via default
        # can now be defined for clouds, but the yaml file only
        # specifies one value for all clouds.

        # get the log level from database
        log_level = Default.get(
            name=LogUtil.LOG_LEVEL_KEY,
            category=LogUtil.category) or LogUtil.DEFAULT_LOG_LEVEL

        # Update the cloudmesh config
        config = ConfigDict("cloudmesh.yaml")
        config["cloudmesh"]["logging"]["level"] = log_level

        # Save this into cloudmesh yaml
        config.save()

    @staticmethod
    def set_level(log_level):
        """
        sets th eloglevel in the database and the loglevel file from
        cloudmesh.yaml
        :param log_level: the loglevel
        :return:
        """
        # TODO: BUG: This seems inconsistent with our use as it mixes db and
        # cloudmesh.yaml.
        level = log_level.upper()

        Default.set(key=LogUtil.LOG_LEVEL_KEY,
                    value=log_level,
                    category=LogUtil.category)

        # get log level obj
        log_level_obj = LogUtil.get_level_obj(log_level)

        # Read the ConfigDict
        config = ConfigDict("cloudmesh.yaml")
        log_file = config["cloudmesh"]["logging"]["file"]

        # Set the logger config
        logging.basicConfig(format=LogUtil.FORMAT,
                            level=log_level_obj,
                            filename=path_expand(log_file))

        LOGGER.info("Set log level to: " + log_level)
        return "Ok."

    @staticmethod
    def get_level():
        """
        get the log level from database
        :param cloudname: The name of the cloud
        :return: the log level
        """
        log_level = Default.get(name=LogUtil.LOG_LEVEL_KEY,
                                category=LogUtil.category)

        LOGGER.info("Returning Log Level: " + log_level)

        # return the level
        return log_level

    @staticmethod
    def initialize_logging():
        """
        reads the log level from the cloudmesh.yaml file from
        cloudmesh.logging.level. If the value is not set the logging will be
        set to the default which is "ERROR"
        :return: the loglevel
        """
        config = ConfigDict("cloudmesh.yaml")
        log_level = config["cloudmesh"]["logging"]["level"] or LogUtil.DEFAULT_LOG_LEVEL

        print("PPPP", log_level)

        # Set the log level
        LogUtil.set_level(log_level)

        return

    @staticmethod
    def get_logger():
        """
        get caller file name
        :return: file name based on the context where the logger is caller
        """

        frame = inspect.stack()[1]
        module = inspect.getmodule(frame[0])
        the_class = module.__name__

        # return logger object
        return logging.getLogger(the_class)

    @staticmethod
    def get_level_obj(log_level):
        """
        gets the log level when passing a string
        :param log_level: case insensitive string. Valid values are debug,
                          info, warning, critical, error
        :return: a logging level
        """
        # Return log level obj
        level = log_level.lower()

        if level == "debug":
            log_level = logging.DEBUG
        elif level == "info":
            level = logging.INFO
        elif level == "warning":
            level = logging.WARNING
        elif level == "critical":
            level = logging.CRITICAL
        elif level == "error":
            level = logging.ERROR
        else:
            level = logging.DEBUG

        return level



from __future__ import print_function

from cloudmesh_client.shell.console import Console
from cloudmesh_client.cloud.iaas.CloudProviderBase import CloudProviderBase

class CloudProviderOpenstackAPI(CloudProviderBase):
    cloud_type = "openstack"

    def __init__(self, cloud_name, cloud_details, user=None):
        super(CloudProviderOpenstackAPI, self).__init__(cloud_name, user=user)

        self.cloud_type = "openstack"
        self.kind = ["image", "flavor", "vm"]

        self.provider = None
        self.default_image = None
        self.default_flavor = None
        self.cloud = None
        self.cloud_details = None

        self.initialize(cloud_name, cloud_details)

    def initialize(self, cloudname, user=None):
        """
        reads the details for the initialization from the cloudname defined in the yaml file
        :param cloudname:
        :param user:
        :return:
        """
        Console.TODO("not yet implemented")

    def list_flavor(self, cloudname, **kwargs):
        Console.TODO("not yet implemented")

    def list_image(self, cloudname, **kwargs):
        Console.TODO("not yet implemented")

    def list_vm(self, cloudname, **kwargs):
        Console.TODO("not yet implemented")

    def list_secgroup_rules(self, cloudname):
        Console.TODO("not yet implemented")

    def list_secgroup(self, cloudname):
        Console.TODO("not yet implemented")

    def boot_vm(self,
                name,
                group=None,
                image=None,
                flavor=None,
                cloud=None,
                key=None,
                secgroup=None,
                meta=None,
                nics=None,
                **kwargs):
        Console.TODO("not yet implemented")

    def delete_vm(self, name, group=None, force=None):
        Console.TODO("not yet implemented")

    def assign_ip(self, name):
        Console.TODO("not yet implemented")

    #
    # All other must methods defined bellow so we can discuss
    #



from __future__ import print_function


def my_i (args):
    print ("i", args)

def my_ii (args):
    print ("ii", args)

def my_iii (args):
    print ("iii", args)

def my_c (cls, args):
    print ("c", args)

def my_cc (cls, args):
    print ("cc", args)

def my_ccc (cls, args):
    print ("ccc", args)


